"""File b21multiple_models_spec.py

:author: Michel Bierlaire, EPFL
:date: Fri Jul 21 17:46:09 2023

 Example of the estimation of several versions of the model using
 assisted specification algorithm. Specification of the catalogs.

"""
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, logzero
from biogeme.catalog import Catalog, segmentation_catalogs

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    MALE,
    INCOME,
    GA,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Segmentations
gender_segmentation = database.generate_segmentation(
    variable=MALE,
    mapping={
        0: 'female',
        1: 'male',
    },
)

income_segmentation = database.generate_segmentation(
    variable=INCOME,
    mapping={
        0: 'inc-zero',
        1: 'inc-under50',
        2: 'inc-50-100',
        3: 'inc-100+',
        4: 'inc-unknown',
    },
)

print(f'{income_segmentation=}')
ga_segmentation = database.generate_segmentation(
    variable=GA, mapping={1: 'GA', 0: 'noGA'}
)

asc_segmentations = (
    gender_segmentation,
    ga_segmentation,
)

ASC_CAR_catalog, ASC_TRAIN_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_CAR, ASC_TRAIN],
    potential_segmentations=asc_segmentations,
    maximum_number=2,
)

cost_segmentations = (
    ga_segmentation,
    income_segmentation,
)

# Note that the function returns a list. In this case, it contains
# only one element. This is the reason of the presence of the comma
# after B_COST_catalog
(B_COST_catalog,) = segmentation_catalogs(
    generic_name='B_COST',
    beta_parameters=[B_COST],
    potential_segmentations=cost_segmentations,
    maximum_number=1,
)

ell_time = Beta('lambda_time', 1, None, 10, 0)
# Potential non linear specification of travel time
TRAIN_TT_catalog = Catalog.from_dict(
    catalog_name='TRAIN_TT',
    dict_of_expressions={
        'linear': TRAIN_TT_SCALED,
        'log': logzero(TRAIN_TT_SCALED),
        'boxcox': models.boxcox(TRAIN_TT_SCALED, ell_time),
    },
)

SM_TT_catalog = Catalog.from_dict(
    catalog_name='SM_TT',
    dict_of_expressions={
        'linear': SM_TT_SCALED,
        'log': logzero(SM_TT_SCALED),
        'boxcox': models.boxcox(SM_TT_SCALED, ell_time),
    },
    controlled_by=TRAIN_TT_catalog.controlled_by,
)

CAR_TT_catalog = Catalog.from_dict(
    catalog_name='CAR_TT',
    dict_of_expressions={
        'linear': CAR_TT_SCALED,
        'log': logzero(CAR_TT_SCALED),
        'boxcox': models.boxcox(CAR_TT_SCALED, ell_time),
    },
    controlled_by=TRAIN_TT_catalog.controlled_by,
)

# Definition of the utility functions with linear cost
V1 = ASC_TRAIN_catalog + B_TIME * TRAIN_TT_catalog + B_COST_catalog * TRAIN_COST_SCALED
V2 = B_TIME * SM_TT_catalog + B_COST_catalog * SM_COST_SCALED
V3 = ASC_CAR_catalog + B_TIME * CAR_TT_catalog + B_COST_catalog * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

print(
    f'Total number of possible specifications: '
    f'{logprob.number_of_multiple_expressions()}'
)

the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b21multiple_models'

PARETO_FILE_NAME = 'b21multiple_models.pareto'



import os
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, Derive

myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

globals().update(database.variables)

exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

MU_EXISTING = Beta('MU_EXISTING', 1, 1, None, 0)
MU_PUBLIC = Beta('MU_PUBLIC', 1, 1, None, 0)
ALPHA_EXISTING = Beta('ALPHA_EXISTING', 0.5, 0, 1, 0)
ALPHA_PUBLIC = 1 - ALPHA_EXISTING


SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

TRAIN_TT_SCALED = TRAIN_TT / 100.0
TRAIN_COST_SCALED = database.DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100
)
SM_TT_SCALED = database.DefineVariable('SM_TT_SCALED', SM_TT / 100.0)
SM_COST_SCALED = database.DefineVariable('SM_COST_SCALED', SM_COST / 100)
CAR_TT_SCALED = database.DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = database.DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


# Associate the availability conditions with the alternatives
CAR_AV_SP = database.DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0))
TRAIN_AV_SP = database.DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0))

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests:
alpha_existing = {1: ALPHA_EXISTING, 2: 0.0, 3: 1.0}

alpha_public = {1: ALPHA_PUBLIC, 2: 1.0, 3: 0.0}

nest_existing = MU_EXISTING, alpha_existing
nest_public = MU_PUBLIC, alpha_public
nests = nest_existing, nest_public

# The choice model is a cross-nested logit, with availability conditions
logprob = models.logcnl_avail(V, av, nests, CHOICE)
prob1 = models.cnl_avail(V, av, nests, 1)
genelas1 = Derive(prob1, 'TRAIN_TT') * TRAIN_TT / prob1
simulate = {'Prob. train': prob1, 'Elas. 1': genelas1}


class test_11(unittest.TestCase):
    def testEstimationAndSimulation(self):
        biogeme = bio.BIOGEME(database, logprob)
        biogeme.saveIterations = False
        biogeme.generateHtml = False
        biogeme.generatePickle = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, -5214.049, 2)
        biosim = bio.BIOGEME(database, simulate)
        simresults = biosim.simulate(results.getBetaValues())
        self.assertAlmostEqual(
            sum(simresults['Prob. train']), 888.3883902853023, 1
        )


if __name__ == '__main__':
    unittest.main()



""" File: b06estimation_integral.py

 Author: Michel Bierlaire, EPFL
 Date: Thu Apr 13 21:03:03 2023

Estimation of a mixtures of logit models where the integral is
calculated using numerical integration.

"""

import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme import models
from biogeme.expressions import Beta, RandomVariable, Integrate, log

from swissmetro import (
    database,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
    CHOICE,
)

ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
B_TIME_RND = B_TIME + B_TIME_S * omega

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
condprob = models.logit(V, av, CHOICE)
prob = Integrate(condprob * density, 'omega')
logprob = log(prob)

the_biogeme = bio.BIOGEME(database, logprob)

the_biogeme.modelName = '06estimationIntegral'

results = the_biogeme.estimate()

# Get the results in a pandas table
print(results.shortSummary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



""" Implementation of the pdf and CDF of common distributions

:author: Michel Bierlaire

:date: Thu Apr 23 12:01:49 2015

"""

from biogeme.expressions import log, exp, Numeric


def normalpdf(x, mu=Numeric(0.0), s=Numeric(1.0)):
    """
    Normal pdf

    Probability density function of a normal distribution

    .. math:: f(x;\\mu, \\sigma) =
        \\frac{1}{\\sigma \\sqrt{2\\pi}} \\exp{-\\frac{(x-\\mu)^2}{2\\sigma^2}}

    :param x: value at which the pdf is evaluated.
    :type x: float or biogeme.expression
    :param mu: location parameter :math:`\\mu` of the Normal distribution.
        Default: 0.
    :type mu: float or biogeme.expression
    :param s: scale parameter :math:`\\sigma` of the Normal distribution.
        Default: 1.
    :type s: float or biogeme.expression

    :note: It is assumed that :math:`\\sigma > 0`, but it is not
        verified by the code.

    :return: value of the Normal pdf.
    :rtype: float or biogeme.expression"""
    d = -(x - mu) * (x - mu)
    n = 2.0 * s * s
    a = d / n
    num = exp(a)
    den = s * 2.506628275
    p = num / den
    return p


def lognormalpdf(x, mu=0.0, s=1.0):
    """
    Log normal pdf

    Probability density function of a log normal distribution

    .. math:: f(x;\\mu, \\sigma) =
              \\frac{1}{x\\sigma \\sqrt{2\\pi}}
              \\exp{-\\frac{(\\ln x-\\mu)^2}{2\\sigma^2}}


    :param x: value at which the pdf is evaluated.
    :type x: float or biogeme.expression
    :param mu: location parameter :math:`\\mu` of the lognormal distribution.
        Default: 0.
    :type mu: float or biogeme.expression
    :param s: scale parameter :math:`\\sigma` of the lognormal distribution.
        Default: 1.
    :type s: float or biogeme.expression

    :note: It is assumed that :math:`\\sigma > 0`, but it is not
        verified by the code.

    :return: value of the lognormal pdf.
    :rtype: float or biogeme.expression

    """
    d = -(log(x) - mu) * (log(x) - mu)
    n = 2.0 * s * s
    a = d / n
    num = exp(a)
    den = x * s * 2.506628275
    p = (x > 0) * num / den
    return p


def uniformpdf(x, a=-1, b=1.0):
    """
    Uniform pdf

    Probability density function of a uniform distribution.

    .. math::  f(x;a, b) = \\left\\{ \\begin{array}{ll}
              \\frac{1}{b-a} & \\text{for } x \\in [a, b] \\\\
              0 & \\text{otherwise}\\end{array} \\right.

    :param x: argument of the pdf
    :type x: float or biogeme.expression
    :param a: lower bound :math:`a` of the distribution. Default: -1.
    :type a: float or biogeme.expression
    :param b: upper bound :math:`b` of the distribution. Default: 1.
    :type b: float or biogeme.expression
    :note: It is assumed that :math:`a < b`, but it is
        not verified by the code.
    :return: value of the uniform pdf.
    :rtype: float or biogeme.expression
 """
    result = (x < a) * 0.0 + (x >= b) * 0.0 + (x >= a) * (x < b) / (b - a)
    return result


def triangularpdf(x, a=-1.0, b=1.0, c=0.0):
    """
    Triangular pdf

    Probability density function of a triangular distribution

    .. math:: f(x;a, b, c) = \\left\\{ \\begin{array}{ll} 0 &
             \\text{if } x < a \\\\\\frac{2(x-a)}{(b-a)(c-a)} &
             \\text{if } a \\leq x < c \\\\\\frac{2(b-x)}{(b-a)(b-c)} &
             \\text{if } c \\leq x < b \\\\0 & \\text{if } x \\geq b.
             \\end{array} \\right.

    :param x: argument of the pdf
    :type x: float or biogeme.expression
    :param a: lower bound :math:`a` of the distribution. Default: -1.
    :type a: float or biogeme.expression
    :param b: upper bound :math:`b` of the distribution. Default: 1.
    :type b: float or biogeme.expression
    :param c: mode :math:`c` of the distribution. Default: 0.
    :type c: float or biogeme.expression
    :note: It is assumed that :math:`a <  c < b`, but it is
        not verified by the code.
    :return: value of the triangular pdf.
    :rtype: float or biogeme.expression

    """
    result = (
        (x < a) * 0.0
        + (x >= b) * 0.0
        + (x >= a)
        * (x < c)
        * 2.0
        * ((x - a) / ((b - a) * (c - a)))
        * (x >= c)
        * (x < b)
        * 2.0
        * (b - x)
        / ((b - a) * (b - c))
    )
    return result


def logisticcdf(x, mu=Numeric(0.0), s=Numeric(1.0)):
    """
    Logistic CDF

    Cumulative distribution function of a logistic distribution

    .. math:: f(x;\\mu, \\sigma) = \\frac{1}
        {1+\\exp\\left(-\\frac{x-\\mu}{\\sigma} \\right)}

    :param x: value at which the CDF is evaluated.
    :type x: float or biogeme.expression
    :param mu: location parameter :math:`\\mu` of the logistic distribution.
        Default: 0.
    :type mu: float or biogeme.expression
    :param s: scale parameter :math:`\\sigma` of the logistic distribution.
        Default: 1.
    :type s: float or biogeme.expression
    :note: It is assumed that :math:`\\sigma > 0`, but it is
        not verified by the code.

    :return: value of the logistic CDF.
    :rtype: float or biogeme.expression

    """
    result = 1.0 / (1.0 + exp(-(x - mu) / s))
    return result



"""File b16panel_discrete_socio_eco.py

:author: Michel Bierlaire, EPFL
:date: Mon Apr 10 12:07:15 2023

 Example of a discrete mixture of logit models, also called latent class model.
 The class membership model includes socio-economic variables.
 The datafile is organized as panel data.
"""

import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models

from biogeme.expressions import (
    Beta,
    Variable,
    bioDraws,
    MonteCarlo,
    log,
    exp,
    bioMultSum,
)
from swissmetro_panel import (
    flat_database,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    INCOME,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b16panel_discrete_socio_eco.py')

# Parameters to be estimated. One version for each latent class.
NUMBER_OF_CLASSES = 2
B_COST = [Beta(f'B_COST_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)]

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME = [Beta(f'B_TIME_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)]

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = [
    Beta(f'B_TIME_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
B_TIME_RND = [
    B_TIME[i] + B_TIME_S[i] * bioDraws(f'B_TIME_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

# We do the same for the constants, to address serial correlation.
ASC_CAR = [
    Beta(f'ASC_CAR_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_CAR_S = [
    Beta(f'ASC_CAR_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_CAR_RND = [
    ASC_CAR[i] + ASC_CAR_S[i] * bioDraws(f'ASC_CAR_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

ASC_TRAIN = [
    Beta(f'ASC_TRAIN_class{i}', 0, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_TRAIN_S = [
    Beta(f'ASC_TRAIN_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_TRAIN_RND = [
    ASC_TRAIN[i] + ASC_TRAIN_S[i] * bioDraws(f'ASC_TRAIN_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

ASC_SM = [Beta(f'ASC_SM_class{i}', 0, None, None, 1) for i in range(NUMBER_OF_CLASSES)]
ASC_SM_S = [
    Beta(f'ASC_SM_S_class{i}', 1, None, None, 0) for i in range(NUMBER_OF_CLASSES)
]
ASC_SM_RND = [
    ASC_SM[i] + ASC_SM_S[i] * bioDraws(f'ASC_SM_RND_class{i}', 'NORMAL_ANTI')
    for i in range(NUMBER_OF_CLASSES)
]

# Parameters for the class membership model
CLASS_CTE = Beta('CLASS_CTE', 0, None, None, 0)
CLASS_INC = Beta('CLASS_INC', 0, None, None, 0)


# In class 0, it is assumed that the time coefficient is zero
B_TIME_RND[0] = 0

# Utility functions
V1 = [
    [
        ASC_TRAIN_RND[i]
        + B_TIME_RND[i] * Variable(f'{t}_TRAIN_TT_SCALED')
        + B_COST[i] * Variable(f'{t}_TRAIN_COST_SCALED')
        for t in range(1, 10)
    ]
    for i in range(NUMBER_OF_CLASSES)
]
V2 = [
    [
        ASC_SM_RND[i]
        + B_TIME_RND[i] * Variable(f'{t}_SM_TT_SCALED')
        + B_COST[i] * Variable(f'{t}_SM_COST_SCALED')
        for t in range(1, 10)
    ]
    for i in range(NUMBER_OF_CLASSES)
]
V3 = [
    [
        ASC_CAR_RND[i]
        + B_TIME_RND[i] * Variable(f'{t}_CAR_TT_SCALED')
        + B_COST[i] * Variable(f'{t}_CAR_CO_SCALED')
        for t in range(1, 10)
    ]
    for i in range(NUMBER_OF_CLASSES)
]
V = [
    [{1: V1[i][t], 2: V2[i][t], 3: V3[i][t]} for t in range(9)]
    for i in range(NUMBER_OF_CLASSES)
]

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a discrete mixture of logit, with availability conditions
# We calculate the conditional probability for each class
prob = [
    exp(
        bioMultSum(
            [models.loglogit(V[i][t], av, Variable(f'{t+1}_CHOICE')) for t in range(9)]
        )
    )
    for i in range(NUMBER_OF_CLASSES)
]

# Class membership model
W = CLASS_CTE + CLASS_INC * INCOME
PROB_class0 = models.logit({0: W, 1: 0}, None, 0)
PROB_class1 = models.logit({0: W, 1: 0}, None, 1)

# Conditional to the random variables, likelihood for the individual.
probIndiv = PROB_class0 * prob[0] + PROB_class1 * prob[1]

# We integrate over the random variables using Monte-Carlo
logprob = log(MonteCarlo(probIndiv))

# Create the Biogeme object
the_biogeme = bio.BIOGEME(flat_database, logprob)
the_biogeme.modelName = 'b16panel_discrete_socio_eco'

# Estimate the parameters.
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



""" Estimate the parameters of the cross-nested logit models with the generated samples

:author: Michel Bierlaire
:date: Sat Apr 15 16:59:43 2023
"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from biogeme.sampling import mev_cnl_sampling

from sample import SAMPLE_SIZE as N_ALT
from choice_data import database

from model import V, log_probability

mu_asian = Beta('mu_asian', 1, 1, None, 0)
mu_downtown = Beta('mu_downtown', 1, 1, None, 0)

# The nests definition is done through the variables in the file. For
# each nest m and each alternative i, there must be a variable m_i
# that is the level of membership of alternative i to nest m (usually
# called the "alpha" membership parameters).

# In this example, we define two nests, one for Asian restaurants, and
# one for restaurants in Downtown. Note that their overlap, so that we
# indeed have a cross-nested logit model.
nests = {'Asian': mu_asian, 'downtown': mu_downtown}

log_gi = mev_cnl_sampling(V, None, log_probability, nests)
logprob = models.logmev(V, log_gi, None, 0)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'cnl'

# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood({i: 1 for i in range(N_ALT)})

# Estimate the parameters
results = the_biogeme.estimate(recycle=True)
print(results.shortSummary())



"""File b09post_processing.py

:author: Michel Bierlaire, EPFL
:date: Thu Jul 20 17:15:37 2023

We consider the model with 432 specifications:
- 3 models
    - logit
    - nested logit with two nests: public and private transportation
    - nested logit with two nests existing and future modes
- 3 functional form for the travel time variables
    - linear specification,
    - Box-Cox transform,
    - power series,
- 2 specification for the cost coefficients:
    - generic
    - alternative specific
- 2 specification for the travel time coefficients:
    - generic
    - alternative specific
- 4 segmentations for the constants:
    - not segmented
    - segmented by GA (yearly subscription to public transport)
    - segmented by luggage
    - segmented both by GA and luggage
-  3 segmentations for the time coefficients:
    - not segmented
    - segmented with first class
    - segmented with trip purpose

This leads to a total of 432 specifications.

After running the assisted specification algorithm, we use post
processing to re-estimate all Pareto optimal models, and display some
information about the algorithm

"""
try: 
    import matplotlib.pyplot as plt
    can_plot = True
except ModuleNotFoundError:
    can_plot = False
import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme.assisted import ParetoPostProcessing

from everything_spec import model_catalog, database

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b08selected_specification')

PARETO_FILE_NAME = 'b07everything_assisted.pareto'

the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'b09post_processing'

post_processing = ParetoPostProcessing(
    biogeme_object=the_biogeme, pareto_file_name=PARETO_FILE_NAME
)

post_processing.reestimate(recycle=True)

post_processing.log_statistics()

if can_plot:
    _ = post_processing.plot()
    plt.show()



"""File b07everything_assisted.py

:author: Michel Bierlaire, EPFL
:date: Sat Jul 15 15:02:20 2023

Investigate various specifications:
- 3 models
    - logit
    - nested logit with two nests: public and private transportation
    - nested logit with two nests existing and future modes
- 3 functional form for the travel time variables
    - linear specification,
    - Box-Cox transform,
    - power series,
- 2 specification for the cost coefficients:
    - generic
    - alternative specific
- 2 specification for the travel time coefficients:
    - generic
    - alternative specific
- 4 segmentations for the constants:
    - not segmented
    - segmented by GA (yearly subscription to public transport)
    - segmented by luggage
    - segmented both by GA and luggage
-  3 segmentations for the time coefficients:
    - not segmented
    - segmented with first class
    - segmented with trip purpose

This leads to a total of 432 specifications.
The algorithm implemented in the AssistedSpecification object is used to
investigate some of these specifications.

"""
import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme.assisted import AssistedSpecification
from biogeme.multiobjectives import loglikelihood_dimension
from everything_spec import model_catalog, database
from results_analysis import report

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b07everything_assisted')

PARETO_FILE_NAME = 'b07everything_assisted.pareto'

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'b07everything'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# Estimate the parameters
assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=loglikelihood_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

non_dominated_models = assisted_specification.run()

report(non_dominated_models)



"""File b01one_latent_regression.py

Measurement equation where the indicators are assumed continuous.
Linear regression.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 16:42:02 2023

"""

import biogeme.logging as blog
import biogeme.biogeme as bio
import biogeme.loglikelihood as ll
from biogeme.expressions import Beta, Elem, bioMultSum
from optima import (
    database,
    age_65_more,
    formulaIncome,
    moreThanOneCar,
    moreThanOneBike,
    individualHouse,
    male,
    haveChildren,
    haveGA,
    highEducation,
    Envir01,
    Envir02,
    Envir03,
    Mobil11,
    Mobil14,
    Mobil16,
    Mobil17,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b01one_latent_regression.py')

# Parameters to be estimated
coef_intercept = Beta('coef_intercept', 0.0, None, None, 0)
coef_age_65_more = Beta('coef_age_65_more', 0.0, None, None, 0)
coef_haveGA = Beta('coef_haveGA', 0.0, None, None, 0)
coef_moreThanOneCar = Beta('coef_moreThanOneCar', 0.0, None, None, 0)
coef_moreThanOneBike = Beta('coef_moreThanOneBike', 0.0, None, None, 0)
coef_individualHouse = Beta('coef_individualHouse', 0.0, None, None, 0)
coef_male = Beta('coef_male', 0.0, None, None, 0)
coef_haveChildren = Beta('coef_haveChildren', 0.0, None, None, 0)
coef_highEducation = Beta('coef_highEducation', 0.0, None, None, 0)


# Latent variable: structural equation

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formulaIncome
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
)


# Measurement equations
INTER_Envir01 = Beta('INTER_Envir01', 0, None, None, 1)
INTER_Envir02 = Beta('INTER_Envir02', 0, None, None, 0)
INTER_Envir03 = Beta('INTER_Envir03', 0, None, None, 0)
INTER_Mobil11 = Beta('INTER_Mobil11', 0, None, None, 0)
INTER_Mobil14 = Beta('INTER_Mobil14', 0, None, None, 0)
INTER_Mobil16 = Beta('INTER_Mobil16', 0, None, None, 0)
INTER_Mobil17 = Beta('INTER_Mobil17', 0, None, None, 0)

B_Envir01_F1 = Beta('B_Envir01_F1', -1, None, None, 1)
B_Envir02_F1 = Beta('B_Envir02_F1', -1, None, None, 0)
B_Envir03_F1 = Beta('B_Envir03_F1', 1, None, None, 0)
B_Mobil11_F1 = Beta('B_Mobil11_F1', 1, None, None, 0)
B_Mobil14_F1 = Beta('B_Mobil14_F1', 1, None, None, 0)
B_Mobil16_F1 = Beta('B_Mobil16_F1', 1, None, None, 0)
B_Mobil17_F1 = Beta('B_Mobil17_F1', 1, None, None, 0)

MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

SIGMA_STAR_Envir01 = Beta('SIGMA_STAR_Envir01', 1, None, None, 0)
SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02', 1, None, None, 0)
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03', 1, None, None, 0)
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11', 1, None, None, 0)
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14', 1, None, None, 0)
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16', 1, None, None, 0)
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17', 1, None, None, 0)

# We build a dict with each contribution to the loglikelihood if
# (var > 0) and (var < 6). If not, 0 is returned.
F = {}
F['Envir01'] = Elem(
    {
        0: 0,
        1: ll.loglikelihoodregression(Envir01, MODEL_Envir01, SIGMA_STAR_Envir01),
    },
    (Envir01 > 0) * (Envir01 < 6),
)
F['Envir02'] = Elem(
    {
        0: 0,
        1: ll.loglikelihoodregression(Envir02, MODEL_Envir02, SIGMA_STAR_Envir02),
    },
    (Envir02 > 0) * (Envir02 < 6),
)
F['Envir03'] = Elem(
    {
        0: 0,
        1: ll.loglikelihoodregression(Envir03, MODEL_Envir03, SIGMA_STAR_Envir03),
    },
    (Envir03 > 0) * (Envir03 < 6),
)
F['Mobil11'] = Elem(
    {
        0: 0,
        1: ll.loglikelihoodregression(Mobil11, MODEL_Mobil11, SIGMA_STAR_Mobil11),
    },
    (Mobil11 > 0) * (Mobil11 < 6),
)
F['Mobil14'] = Elem(
    {
        0: 0,
        1: ll.loglikelihoodregression(Mobil14, MODEL_Mobil14, SIGMA_STAR_Mobil14),
    },
    (Mobil14 > 0) * (Mobil14 < 6),
)
F['Mobil16'] = Elem(
    {
        0: 0,
        1: ll.loglikelihoodregression(Mobil16, MODEL_Mobil16, SIGMA_STAR_Mobil16),
    },
    (Mobil16 > 0) * (Mobil16 < 6),
)
F['Mobil17'] = Elem(
    {
        0: 0,
        1: ll.loglikelihoodregression(Mobil17, MODEL_Mobil17, SIGMA_STAR_Mobil17),
    },
    (Mobil17 > 0) * (Mobil17 < 6),
)

# The log likelihood is the sum of the elements of the above dict
loglike = bioMultSum(F)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, loglike)
the_biogeme.modelName = 'b01one_latent_regression'

# Estimate the parameters
results = the_biogeme.estimate()

print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')
results.writeLaTeX()
print(f'LaTeX file: {results.data.latexFileName}')



"""File b08arc_elasticities.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:00:10 2023

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We calculate disaggregate and aggregate direct arc elasticities.
"""

import sys
import biogeme.biogeme as bio
import biogeme.exceptions as excep
from biogeme import models
import biogeme.results as res
from optima_data import database, normalized_weight
from scenarios import scenario

# Obtain the specification for the default scenario
V, nests, _, MarginalCostPT = scenario()

# Obtain the expression for the choice probability of each alternative
prob_PT = models.nested(V, None, nests, 0)

# We investigate a scenario where the price for public transportation
# increases by 20%
V_after, _, _, MarginalCostPT_after = scenario(factor=1.2)
prob_PT_after = models.nested(V_after, None, nests, 0)


# Disaggregate elasticities
direct_elas_pt = (
    (prob_PT_after - prob_PT)
    * MarginalCostPT
    / (prob_PT * (MarginalCostPT_after - MarginalCostPT))
)

simulate = {
    'weight': normalized_weight,
    'Prob. PT': prob_PT,
    'direct_elas_pt': direct_elas_pt,
}

the_biogeme = bio.BIOGEME(database, simulate)

# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02estimation.py in order to generate '
        'the file b02estimation.pickle.'
    )

# simulated_values is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulated_values = the_biogeme.simulate(results.getBetaValues())

# We calculate the elasticities
simulated_values['Weighted prob. PT'] = (
    simulated_values['weight'] * simulated_values['Prob. PT']
)

denominator_pt = simulated_values['Weighted prob. PT'].sum()

direct_elas_pt = (
    simulated_values['Weighted prob. PT']
    * simulated_values['direct_elas_pt']
    / denominator_pt
).sum()

print(
    f'Aggregate direct arc elasticity of public transportation wrt cost: '
    f'{direct_elas_pt:.3g}'
)



import unittest
from biogeme.expressions import Beta, Variable
from biogeme.exceptions import BiogemeError
from biogeme.segmentation import DiscreteSegmentationTuple
from biogeme.catalog import (
    Catalog,
    SegmentedParameters,
    segmentation_catalogs,
    generic_alt_specific_catalogs,
)
from biogeme.multiple_expressions import NamedExpression
from biogeme.configuration import (
    SelectionTuple,
    Configuration,
)


class TestCatalog(unittest.TestCase):
    def setUp(self):
        self.expression1 = Beta('expr1', 0, None, None, 0)
        self.expression2 = Beta('expr2', 0, None, None, 0)
        self.expression3 = Beta('expr3', 0, None, None, 0)

        self.named_expression1 = NamedExpression('expression1', self.expression1)
        self.named_expression2 = NamedExpression('expression2', self.expression2)
        self.named_expression3 = NamedExpression('expression3', self.expression3)

        self.catalog = Catalog(
            'my_catalog',
            [self.named_expression1, self.named_expression2, self.named_expression3],
        )

        self.another_catalog = Catalog(
            'another_catalog',
            [self.named_expression1, self.named_expression2, self.named_expression3],
        )

    def test_catalog_creation(self):
        # Test catalog creation with valid input
        catalog = Catalog(
            'test_catalog', [self.named_expression1, self.named_expression2]
        )

        self.assertEqual(catalog.name, 'test_catalog')
        self.assertEqual(len(catalog.named_expressions), 2)
        self.assertEqual(catalog.catalog_size(), 2)
        self.assertEqual(catalog.named_expressions[0].name, 'expression1')
        self.assertEqual(catalog.named_expressions[1].name, 'expression2')
        self.assertIs(catalog.named_expressions[0].expression, self.expression1)
        self.assertIs(catalog.named_expressions[1].expression, self.expression2)

        # Test catalog creation with an empty list
        with self.assertRaises(BiogemeError):
            Catalog('empty_catalog', [])

        # Trying to control with a relevant Catalog
        other_catalog = Catalog(
            catalog_name='other_catalog',
            named_expressions=[self.named_expression1, self.named_expression2],
            controlled_by=catalog.controlled_by,
        )
        self.assertIs(other_catalog.controlled_by, catalog.controlled_by)

        # Trying to control with an irrelevant Catalog
        with self.assertRaises(BiogemeError):
            other_catalog = Catalog(
                catalog_name='other_catalog',
                named_expressions=[self.named_expression1, self.named_expression2],
                controlled_by=self.catalog.controlled_by,
            )

    def test_catalog_from_dict(self):
        # Test catalog creation from a dictionary
        dict_of_expressions = {'expr1': self.expression1, 'expr2': self.expression2}
        catalog = Catalog.from_dict('test_catalog', dict_of_expressions)

        self.assertEqual(catalog.name, 'test_catalog')
        self.assertEqual(len(catalog.named_expressions), 2)
        self.assertEqual(catalog.named_expressions[0].name, 'expr1')
        self.assertEqual(catalog.named_expressions[1].name, 'expr2')
        self.assertIs(catalog.named_expressions[0].expression, self.expression1)
        self.assertIs(catalog.named_expressions[1].expression, self.expression2)
        self.assertEqual(catalog.catalog_size(), 2)

    def test_set_of_configurations(self):
        config_1 = Configuration(
            [SelectionTuple(controller='my_catalog', selection='expression1')]
        )
        config_2 = Configuration(
            [SelectionTuple(controller='my_catalog', selection='expression2')]
        )
        config_3 = Configuration(
            [SelectionTuple(controller='my_catalog', selection='expression3')]
        )
        expected_set = {
            config_1,
            config_2,
            config_3,
        }
        the_set = self.catalog.set_of_configurations()
        self.assertSetEqual(the_set, expected_set)

    def test_selected_expression(self):
        selected_expression = self.catalog.selected_expression()
        self.assertEqual(selected_expression, self.expression1)

    def test_selected_name(self):
        selected_name = self.catalog.selected_name()
        self.assertEqual(selected_name, 'expression1')

    def test_current_configuration(self):
        configuration = self.catalog.current_configuration()
        self.assertEqual(len(configuration.selections), 1)
        self.assertEqual(configuration.selections[0].controller, 'my_catalog')
        self.assertEqual(configuration.selections[0].selection, 'expression1')

    def test_iterator(self):
        my_expression = self.catalog + self.another_catalog
        configurations = {e.current_configuration() for e in my_expression}
        expected_configurations = {
            Configuration.from_string(
                'another_catalog:expression1;my_catalog:expression1'
            ),
            Configuration.from_string(
                'another_catalog:expression1;my_catalog:expression2'
            ),
            Configuration.from_string(
                'another_catalog:expression1;my_catalog:expression3'
            ),
            Configuration.from_string(
                'another_catalog:expression2;my_catalog:expression1'
            ),
            Configuration.from_string(
                'another_catalog:expression2;my_catalog:expression2'
            ),
            Configuration.from_string(
                'another_catalog:expression2;my_catalog:expression3'
            ),
            Configuration.from_string(
                'another_catalog:expression3;my_catalog:expression1'
            ),
            Configuration.from_string(
                'another_catalog:expression3;my_catalog:expression2'
            ),
            Configuration.from_string(
                'another_catalog:expression3;my_catalog:expression3'
            ),
        }
        self.assertSetEqual(configurations, expected_configurations)


class SegmentationCatalogTest(unittest.TestCase):
    def setUp(self):
        self.beta_parameter_1 = Beta('beta_1', 0, None, 10, 0)
        self.beta_parameter_2 = Beta('beta_2', 0.5, -10, 10, 0)
        self.variable_x = Variable('x')
        self.variable_y = Variable('y')
        self.variable_z = Variable('z')
        self.segmentation1 = DiscreteSegmentationTuple(
            self.variable_x, {0: 'zero', 1: 'one'}
        )
        self.segmentation2 = DiscreteSegmentationTuple(
            self.variable_y, {0: 'zero', 1: 'one'}
        )
        self.segmentation3 = DiscreteSegmentationTuple(
            self.variable_z, {0: 'zero', 1: 'one'}
        )
        self.potential_segmentations = (
            self.segmentation1,
            self.segmentation2,
            self.segmentation3,
        )

    def test_wrong_segment_name(self):
        wrong_segmentation_1 = DiscreteSegmentationTuple(
            self.variable_z, {0: 'zero:', 1: 'one'}
        )
        potential_segmentations_1 = (
            wrong_segmentation_1,
        )
        with self.assertRaises(BiogemeError):
            _ = segmentation_catalogs(
                generic_name='wrong_segmentation',
                beta_parameters=[self.beta_parameter_1],
                potential_segmentations=potential_segmentations_1,
                maximum_number=1,
            )
            
        
        wrong_segmentation_2 = DiscreteSegmentationTuple(
            self.variable_z, {0: 'zero', 1: 'on;e'}
        )
        potential_segmentations_2 = (
            wrong_segmentation_2,
        )
        with self.assertRaises(BiogemeError):
            _ = segmentation_catalogs(
                generic_name='wrong_segmentation',
                beta_parameters=[self.beta_parameter_1],
                potential_segmentations=potential_segmentations_2,
                maximum_number=1,
            )


    def test_segmentation_catalog_no_sync(self):
        catalogs = segmentation_catalogs(
            generic_name='segmentated_beta',
            beta_parameters=[self.beta_parameter_1],
            potential_segmentations=self.potential_segmentations,
            maximum_number=2,
        )
        self.assertIsInstance(catalogs[0], Catalog)
        self.assertEqual(catalogs[0].name, 'segmented_beta_1')
        self.assertEqual(len(catalogs[0].named_expressions), 7)

    def test_segmentation_catalog_with_sync(self):
        catalogs = segmentation_catalogs(
            generic_name='segmented_beta',
            beta_parameters=[self.beta_parameter_1, self.beta_parameter_2],
            potential_segmentations=self.potential_segmentations,
            maximum_number=2,
        )
        self.assertIsInstance(catalogs[0], Catalog)
        self.assertIsInstance(catalogs[1], Catalog)
        self.assertEqual(catalogs[0].name, 'segmented_beta_1')
        self.assertEqual(catalogs[1].name, 'segmented_beta_2')
        self.assertEqual(len(catalogs[0].named_expressions), 7)
        self.assertIs(catalogs[0].controlled_by, catalogs[1].controlled_by)

    def test_segmentation_catalog_no_segmentations(self):
        catalogs = segmentation_catalogs(
            generic_name='segmented_beta',
            beta_parameters=[self.beta_parameter_1],
            potential_segmentations=(),
            maximum_number=2,
        )
        self.assertIsInstance(catalogs[0], Catalog)
        self.assertEqual(catalogs[0].name, 'segmented_beta_1')
        self.assertEqual(len(catalogs[0].named_expressions), 1)


class TestSegmentedParameters(unittest.TestCase):
    def setUp(self):
        self.beta_parameters = [
            Beta('beta_time', 0, None, None, 0),
            Beta('beta_cost', 0, None, None, 0),
        ]
        self.alternatives = ['train', 'car', 'sm']
        self.segmented_parameters = SegmentedParameters(
            beta_parameters=self.beta_parameters,
            alternatives=self.alternatives,
        )

    def test_constructor(self):
        self.assertEqual(len(self.segmented_parameters.all_parameters), 2 * 4)
        self.assertEqual(len(self.segmented_parameters.beta_parameters), 2)
        self.assertEqual(len(self.segmented_parameters.alternatives), 3)


class TestGenericAltSpecificCatalog(unittest.TestCase):
    def test_generic_alt_specific_catalog(self):
        coefficient = Beta(
            name='coefficient', value=1.0, lowerbound=None, upperbound=None, status=0
        )
        alternatives = ('A', 'B', 'C')

        result = generic_alt_specific_catalogs(
            generic_name='coefficient',
            beta_parameters=[coefficient],
            alternatives=alternatives
        )

        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], dict)
        self.assertEqual(len(result[0]), len(alternatives))

        for alternative, catalog in result[0].items():
            self.assertIsInstance(catalog, Catalog)
            for named_expression in catalog.named_expressions:
                if named_expression.name == 'generic':
                    self.assertEqual(named_expression.expression, coefficient)
                elif named_expression.name == 'altspec':
                    beta = named_expression.expression
                    self.assertIsInstance(beta, Beta)
                    self.assertEqual(beta.name, f'{coefficient.name}_{alternative}')
                    self.assertEqual(beta.initValue, coefficient.initValue)
                    self.assertEqual(beta.lb, coefficient.lb)
                    self.assertEqual(beta.ub, coefficient.ub)
                    self.assertEqual(beta.status, coefficient.status)
                else:
                    self.fail(f'Unknown expression {named_expression.name}')

    def test_generic_alt_specific_catalog_invalid_coefficient(self):
        coefficient = 1.0  # Invalid coefficient, not a Beta object
        alternatives = ('A', 'B', 'C')

        with self.assertRaises(BiogemeError):
            generic_alt_specific_catalogs(
                generic_name='any',
                beta_parameters=[coefficient],
                alternatives=alternatives,
            )

    def test_generic_alt_specific_catalog_insufficient_alternatives(self):
        coefficient = Beta(
            name='coefficient', value=1.0, lowerbound=None, upperbound=None, status=0
        )
        alternatives = ('A',)

        with self.assertRaises(BiogemeError):
            generic_alt_specific_catalogs(
                generic_name='any',
                beta_parameters=[coefficient], alternatives=alternatives
            )


if __name__ == '__main__':
    unittest.main()



"""File knapsack.py

:author: Michel Bierlaire
:date: Fri Mar 17 08:24:33 2023

This illustrates how to use the multi-objective VNS algorithm using a
simple knapsack problem.  There are two objectives: maximize utility
and minimize weight. The weight cannot go beyond capacity.

"""
import random
import numpy as np
from biogeme import vns

from biogeme.pareto import SetElement


# pylint: disable=invalid-name


class Sack:
    """Implements a solution. Here, a sack configuration."""

    SEPARATOR = '-'
    utility_data = None
    weight_data = None

    def __init__(self, string_representation):
        """Creates a sack from a string representation"""
        self.decisions = [int(i) for i in string_representation.split(self.SEPARATOR)]
        self.utility = sum(x * u for x, u in zip(self.decisions, self.utility_data))
        self.weight = sum(x * u for x, u in zip(self.decisions, self.weight_data))

    @classmethod
    def from_decisions(cls, decisions):
        """Creates a sack from the actual decisions"""
        the_string = cls.code_decisions(decisions)
        return cls(the_string)

    def __str__(self):
        return self.describe()

    def __repr__(self):
        return self.code_id()

    @classmethod
    def empty_sack(cls, size):
        """
        :return: empty sack of the same dimension.
        :rtype: class Sack
        """
        element_id = cls.SEPARATOR.join(['0'] * size)
        return cls(element_id)

    def get_element(self):
        """Implementation of abstract method"""
        return SetElement(self.code_id(), [-self.utility, self.weight])

    def describe(self):
        """Short description of the solution. Used for reporting.

        :return: short description of the solution.
        :rtype: str
        """
        return f'{self.code_id()}: U={self.utility} W={self.weight}'

    def code_id(self):
        """Provide a string ID for the sack

        :return: identifier of the solution. Used to organize the Pareto set.
        :rtype: str
        """
        return self.code_decisions(self.decisions)

    @classmethod
    def code_decisions(cls, decisions):
        """Provide a string ID for the sack given the decisions

        :return: identifier of the solution. Used to organize the Pareto set.
        :rtype: str
        """
        return cls.SEPARATOR.join([str(x) for x in decisions])


# Operators
def add_items(element, size=1):
    """Add ``size`` items in the sack

    :param element: ID of the current sack
    :type element: class SetElement

    :param size: number of items to add into the sack
    :type size: int

    :return: Id of the new sack, and number of changes actually made
    :rtype: tuple(class SetElement, int)
    """
    solution = Sack(element.element_id)
    absent = [i for i, x in enumerate(solution.decisions) if x == 0]
    if not absent:
        return None, 0
    random.shuffle(absent)
    n = min(len(absent), size)
    xplus = solution.decisions.copy()
    for i in absent[:n]:
        xplus[i] = 1
    neighbor = Sack.from_decisions(xplus)
    return neighbor.get_element(), n


def remove_items(element, size=1):
    """Remove ``size`` items from the sack

    :param element: current sack
    :type element: class Sack

    :param size: number of items to remove from the sack
    :type size: int

    :return: Id of the new sack, and number of changes actually made
    :rtype: tuple(class SetElement, int)
    """
    solution = Sack(element.element_id)
    present = [i for i, x in enumerate(solution.decisions) if x == 1]
    if not present:
        return None, 0
    random.shuffle(present)
    n = min(len(present), size)
    xplus = solution.decisions.copy()
    for i in present[:n]:
        xplus[i] = 0
    neighbor = Sack.from_decisions(xplus)
    return neighbor.get_element(), n


def change_decisions(element, size=1):
    """Change the status of ``size`` items in the sack.

    :param element: representation of the current sack
    :type element: class SetElement

    :param size: number of items to modify
    :type size: int

    :return: representation of the new sack, and number of changes actually made
    :rtype: tuple(class SetElement, int)

    """
    solution = Sack(element.element_id)
    length = len(solution.decisions)
    n = min(length, size)
    order = np.random.permutation(length)
    xplus = solution.decisions.copy()
    for i in range(n):
        xplus[order[i]] = 1 - xplus[order[i]]
    neighbor = Sack.from_decisions(xplus)
    return neighbor.get_element(), n


class Knapsack(vns.ProblemClass):
    """Class defining the knapsack problem. Note the inheritance from the
    abstract class used by the VNS algorithm. It guarantees the
    compliance with the requirements of the algorithm.

    """

    def __init__(self, utility, weight, capacity):
        """Ctor"""
        self.utility = utility
        self.weight = weight
        self.capacity = capacity
        self.operators = {
            'Add items': add_items,
            'Remove items': remove_items,
            'Change decision for items': change_decisions,
        }
        self.currentSolution = None
        self.lastOperator = None
        super().__init__(self.operators)

    def is_valid(self, element):
        """Check if the sack verifies the capacity constraint

        Implementation of the abstract method

        :param element: representation of the sack to check
        :type element: class SetElement

        :return: True if the capacity constraint is verified
        :rtype: bool
        """
        solution = Sack(element.element_id)
        if solution.weight <= self.capacity:
            return True, None
        return False, 'Infeasible sack'



"""File m03_simultaneous_estimation.py

Choice model with the latent variable.
Mixture of logit.
Measurement equation for the indicators.
Maximum likelihood (full information) estimation.

:author: Michel Bierlaire, EPFL
:date: Fri Apr 14 10:07:43 2023
"""

import sys
import biogeme.logging as blog
import biogeme.biogeme as bio
import biogeme.distributions as dist
import biogeme.results as res
from biogeme import models
from biogeme.expressions import (
    Beta,
    Variable,
    log,
    RandomVariable,
    Integrate,
    Elem,
    bioNormalCdf,
    exp,
)

from optima import (
    database,
    male,
    age,
    haveChildren,
    highEducation,
    childCenter,
    childSuburb,
    SocioProfCat,
    TimePT,
    TimeCar,
    MarginalCostPT,
    CostCarCHF,
    distance_km,
    TripPurpose,
    WaitingTimePT,
    Choice,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example m03_simultaneous_estimation.py')


# Read the estimates from the structural equation estimation
NAME = 'm01_latent_variable'
try:
    structResults = res.bioResults(pickleFile=f'{NAME}.pickle')
except FileNotFoundError:
    print(
        f'Run first the script {NAME}.py in order to generate the file '
        f'{NAME}.pickle.'
    )
    sys.exit()
struct_betas = structResults.getBetaValues()


# Coefficients

coef_intercept = Beta('coef_intercept', struct_betas['coef_intercept'], None, None, 0)
coef_age_30_less = Beta(
    'coef_age_30_less', struct_betas['coef_age_30_less'], None, None, 0
)
coef_haveChildren = Beta(
    'coef_haveChildren', struct_betas['coef_haveChildren'], None, None, 0
)

coef_highEducation = Beta(
    'coef_highEducation', struct_betas['coef_highEducation'], None, None, 0
)

coef_artisans = Beta('coef_artisans', struct_betas['coef_artisans'], None, None, 0)

coef_employees = Beta('coef_employees', struct_betas['coef_employees'], None, None, 0)

coef_male = Beta('coef_male', struct_betas['coef_male'], None, None, 0)

coef_child_center = Beta(
    'coef_child_center', struct_betas['coef_child_center'], None, None, 0
)
coef_child_suburb = Beta(
    'coef_child_suburb', struct_betas['coef_child_suburb'], None, None, 0
)

# Latent variable: structural equation

# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

ACTIVELIFE = (
    coef_intercept
    + coef_child_center * childCenter
    + coef_child_suburb * childSuburb
    + coef_highEducation * highEducation
    + coef_artisans * (SocioProfCat == 5)
    + coef_employees * (SocioProfCat == 6)
    + coef_age_30_less * (age <= 30)
    + coef_male * male
    + coef_haveChildren * haveChildren
    + sigma_s * omega
)

# Measurement equations

indicators = [
    'ResidCh01',
    'ResidCh04',
    'ResidCh05',
    'ResidCh06',
    'LifSty07',
    'LifSty10',
]

# We define the intercept parameters. The first one is normalized to 0.
INTER = {
    k: Beta(f'INTER_{k}', struct_betas[f'INTER_{k}'], None, None, 0)
    for k in indicators[1:]
}
INTER[indicators[0]] = Beta(f'INTER_{indicators[0]}', 0, None, None, 1)

# We define the coefficients. The first one is normalized to 1.
B = {k: Beta(f'B_{k}', struct_betas[f'B_{k}'], None, None, 0) for k in indicators[1:]}
B[indicators[0]] = Beta(f'B_{indicators[0]}', 1, None, None, 1)

# We define the measurement equations for each indicator
MODEL = {k: INTER[k] + B[k] * ACTIVELIFE for k in indicators}

# We define the scale parameters of the error terms.
SIGMA_STAR = {
    k: Beta(f'SIGMA_STAR_{k}', struct_betas[f'SIGMA_STAR_{k}'], 1.0e-5, None, 0)
    for k in indicators[1:]
}
SIGMA_STAR[indicators[0]] = Beta(f'SIGMA_STAR_{indicators[0]}', 1, None, None, 1)

delta_1 = Beta('delta_1', struct_betas['delta_1'], 1.0e-5, None, 0)
delta_2 = Beta('delta_2', struct_betas['delta_2'], 1.0e-5, None, 0)
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1
tau_3 = delta_1
tau_4 = delta_1 + delta_2

tau_1_residual = {k: (tau_1 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
tau_2_residual = {k: (tau_2 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
tau_3_residual = {k: (tau_3 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
tau_4_residual = {k: (tau_4 - MODEL[k]) / SIGMA_STAR[k] for k in indicators}
Ind = {
    k: {
        1: bioNormalCdf(tau_1_residual[k]),
        2: bioNormalCdf(tau_2_residual[k]) - bioNormalCdf(tau_1_residual[k]),
        3: bioNormalCdf(tau_3_residual[k]) - bioNormalCdf(tau_2_residual[k]),
        4: bioNormalCdf(tau_4_residual[k]) - bioNormalCdf(tau_3_residual[k]),
        5: 1 - bioNormalCdf(tau_4_residual[k]),
        6: 1.0,
        -1: 1.0,
        -2: 1.0,
    }
    for k in indicators
}

prob_indicators = Elem(Ind[indicators[0]], Variable(indicators[0]))
for k in indicators[1:]:
    prob_indicators *= Elem(Ind[k], Variable(k))

# Choice model
# Read the estimates from the sequential estimation, and use
# them as starting values

NAME_SEQ = 'm02_sequential_estimation'
try:
    choiceResults = res.bioResults(pickleFile=f'{NAME_SEQ}.pickle')
except FileNotFoundError:
    print(
        f'Run first the script {NAME_SEQ}.py in order to generate the file '
        f'{NAME_SEQ}.pickle.'
    )
    sys.exit()

choiceBetas = choiceResults.getBetaValues()

ASC_CAR = Beta('ASC_CAR', choiceBetas['ASC_CAR'], None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', choiceBetas['ASC_SM'], None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', choiceBetas['BETA_COST_HWH'], None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', choiceBetas['BETA_COST_OTHER'], None, None, 0)
BETA_DIST = Beta('BETA_DIST', choiceBetas['BETA_DIST'], None, None, 0)
BETA_TIME_CAR_REF = Beta(
    'BETA_TIME_CAR_REF', choiceBetas['BETA_TIME_CAR_REF'], None, 0, 0
)
BETA_TIME_CAR_AL = Beta(
    'BETA_TIME_CAR_AL', choiceBetas['BETA_TIME_CAR_AL'], None, None, 0
)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', choiceBetas['BETA_TIME_PT_REF'], None, 0, 0)
BETA_TIME_PT_AL = Beta('BETA_TIME_PT_AL', choiceBetas['BETA_TIME_PT_AL'], None, None, 0)
BETA_WAITING_TIME = Beta(
    'BETA_WAITING_TIME', choiceBetas['BETA_WAITING_TIME'], None, None, 0
)

TimePT_scaled = database.DefineVariable('TimePT_scaled', TimePT / 200)
TimeCar_scaled = database.DefineVariable('TimeCar_scaled', TimeCar / 200)
MarginalCostPT_scaled = database.DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10
)
CostCarCHF_scaled = database.DefineVariable('CostCarCHF_scaled', CostCarCHF / 10)
distance_km_scaled = database.DefineVariable('distance_km_scaled', distance_km / 5)
PurpHWH = database.DefineVariable('PurpHWH', TripPurpose == 1)
PurpOther = database.DefineVariable('PurpOther', TripPurpose != 1)

# Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_AL * ACTIVELIFE)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_AL * ACTIVELIFE)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0, 1: V1, 2: V2}

# Conditional to omega, we have a logit model (called the kernel) for
# the choice
condprob = models.logit(V, None, Choice)

# Conditional to omega, we have the product of ordered probit for the
# indicators.
condlike = prob_indicators * condprob

# We integrate over omega using numerical integration
loglike = log(Integrate(condlike * density, 'omega'))


# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, loglike)
the_biogeme.modelName = 'm03_simultaneous_estimation'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.shortSummary())
print(results.getEstimatedParameters())

print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')



import os
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import (
    Beta,
    log,
    exp,
    bioDraws,
    MonteCarlo,
    bioMultSum,
    Variable,
)


myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)
database.panel('ID')

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

globals().update(database.variables)

exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)


SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

TRAIN_TT_SCALED = database.DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = database.DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100
)
SM_TT_SCALED = database.DefineVariable('SM_TT_SCALED', SM_TT / 100.0)
SM_COST_SCALED = database.DefineVariable('SM_COST_SCALED', SM_COST / 100)
CAR_TT_SCALED = database.DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = database.DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

# Associate the availability conditions with the alternatives

CAR_AV_SP = database.DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0))
TRAIN_AV_SP = database.DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0))


flat_df = database.generateFlatPanelDataframe(identical_columns=None)
flat_database = db.Database('swissmetro_flat', flat_df)

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation
SIGMA_CAR = Beta('SIGMA_CAR', 3.7, None, None, 0)
SIGMA_SM = Beta('SIGMA_SM', 0.759, None, None, 0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN', 3.02, None, None, 0)

EC_CAR = SIGMA_CAR * bioDraws('EC_CAR', 'NORMAL')
EC_SM = SIGMA_SM * bioDraws('EC_SM', 'NORMAL')
EC_TRAIN = SIGMA_TRAIN * bioDraws('EC_TRAIN', 'NORMAL')

ASC_CAR = Beta('ASC_CAR', 0.136, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -1, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', -6.3, None, 0, 0)
B_COST = Beta('B_COST', -3.29, None, 0, 0)


# For latent class 1, whete the time coefficient is zero
V11 = [
    ASC_TRAIN + B_COST * Variable(f'{t}_TRAIN_COST_SCALED') + EC_TRAIN
    for t in range(1, 10)
]
V12 = [
    ASC_SM + B_COST * Variable(f'{t}_SM_COST_SCALED') + EC_SM
    for t in range(1, 10)
]
V13 = [
    ASC_CAR + B_COST * Variable(f'{t}_CAR_CO_SCALED') + EC_CAR
    for t in range(1, 10)
]

V1 = [{1: V11[t], 2: V12[t], 3: V13[t]} for t in range(9)]

# For latent class 2, whete the time coefficient is estimated
V21 = [
    ASC_TRAIN
    + B_TIME * Variable(f'{t}_TRAIN_TT_SCALED')
    + B_COST * Variable(f'{t}_TRAIN_COST_SCALED')
    + EC_TRAIN
    for t in range(1, 10)
]
V22 = [
    ASC_SM
    + B_TIME * Variable(f'{t}_SM_TT_SCALED')
    + B_COST * Variable(f'{t}_SM_COST_SCALED')
    + EC_SM
    for t in range(1, 10)
]
V23 = [
    ASC_CAR
    + B_TIME * Variable(f'{t}_CAR_TT_SCALED')
    + B_COST * Variable(f'{t}_CAR_CO_SCALED')
    + EC_CAR
    for t in range(1, 10)
]

V2 = [{1: V21[t], 2: V22[t], 3: V23[t]} for t in range(9)]

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


# Class membership model
# Class membership model
CLASS_CTE = Beta('CLASS_CTE', 0, None, None, 0)
CLASS_INC = Beta('CLASS_INC', 0, None, None, 0)
W1 = CLASS_CTE + CLASS_INC * INCOME
probClass1 = models.logit({1: W1, 2: 0}, None, 1)
probClass2 = models.logit({1: W1, 2: 0}, None, 2)

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 1
obsprob1 = [
    models.loglogit(V1[t], av, Variable(f'{t+1}_CHOICE')) for t in range(9)
]
prob1 = exp(bioMultSum(obsprob1))

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 2
obsprob2 = [
    models.loglogit(V2[t], av, Variable(f'{t+1}_CHOICE')) for t in range(9)
]
prob2 = exp(bioMultSum(obsprob2))

# Conditional to the random variables, likelihood for the individual.
probIndiv = probClass1 * prob1 + probClass2 * prob2

# We integrate over the random variables using Monte-Carlo
logprob = log(MonteCarlo(probIndiv))


class test_16(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(flat_database, logprob, numberOfDraws=5, seed=10)
        biogeme.saveIterations = False
        biogeme.generateHtml = False
        biogeme.generatePickle = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, -4071.8391253004647, 2)


if __name__ == '__main__':
    unittest.main()



"""File b10nested_bottom.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:05:04 2023

 Example of a nested logit model where the normalization is done at
 the bottom level.
"""

import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b10nested_bottom.py')

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# If the lower bound is set to zero, the model cannot be evaluated.
# Therefore, we set the lower bound to a small number, strictly larger
# than zero.
MU = Beta('MU', 0.5, 0.000001, 1.0, 0)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests:
# 1: nests parameter
# 2: list of alternatives
existing = 1.0, [1, 3]
future = 1.0, [2]
nests = existing, future

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
# The choice model is a nested logit, with availability conditions,
# where the scale parameter mu is explicitly involved.
logprob = models.lognestedMevMu(V, av, nests, CHOICE, MU)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b10nested_bottom'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File optima_data.py

Prepare data for the Optima case study
:author: Michel Bierlaire
:date: Wed Apr 12 20:52:37 2023

"""

import pandas as pd
import biogeme.database as db
from biogeme.expressions import Variable


# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# Variables from the data
Choice = Variable('Choice')
TimePT = Variable('TimePT')
TimeCar = Variable('TimeCar')
MarginalCostPT = Variable('MarginalCostPT')
CostCarCHF = Variable('CostCarCHF')
distance_km = Variable('distance_km')
Gender = Variable('Gender')
OccupStat = Variable('OccupStat')
Weight = Variable('Weight')

# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Normalize the weights
sum_weight = database.data['Weight'].sum()
number_of_rows = database.data.shape[0]
normalized_weight = Weight * number_of_rows / sum_weight



"""File b07problem_simul.py

 This file is an updated version of 07problem.py, where
 the probabilities are simulated in order to
 investigate the numerical issue.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 18:21:15 2023

"""

import biogeme.biogeme as bio
import biogeme.logging as blog
from biogeme.expressions import Beta, Elem, bioNormalCdf

from optima import (
    database,
    age_65_more,
    formulaIncome,
    moreThanOneCar,
    moreThanOneBike,
    individualHouse,
    male,
    haveChildren,
    haveGA,
    highEducation,
    Envir01,
    Envir02,
    Envir03,
    Mobil11,
    Mobil14,
    Mobil16,
    Mobil17,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b07problem_simul.py')

# Parameters to be estimated
coef_intercept = Beta('coef_intercept', 0.0, None, None, 0)
coef_age_65_more = Beta('coef_age_65_more', 0.0, None, None, 0)
coef_haveGA = Beta('coef_haveGA', 0.0, None, None, 0)
coef_moreThanOneCar = Beta('coef_moreThanOneCar', 0.0, None, None, 0)
coef_moreThanOneBike = Beta('coef_moreThanOneBike', 0.0, None, None, 0)
coef_individualHouse = Beta('coef_individualHouse', 0.0, None, None, 0)
coef_male = Beta('coef_male', 0.0, None, None, 0)
coef_haveChildren = Beta('coef_haveChildren', 0.0, None, None, 0)
coef_highEducation = Beta('coef_highEducation', 0.0, None, None, 0)

# Latent variable: structural equation

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formulaIncome
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
)

# Measurement equations

INTER_Envir01 = Beta('INTER_Envir01', 0, None, None, 1)
INTER_Envir02 = Beta('INTER_Envir02', 0, None, None, 0)
INTER_Envir03 = Beta('INTER_Envir03', 0, None, None, 0)
INTER_Mobil11 = Beta('INTER_Mobil11', 0, None, None, 0)
INTER_Mobil14 = Beta('INTER_Mobil14', 0, None, None, 0)
INTER_Mobil16 = Beta('INTER_Mobil16', 0, None, None, 0)
INTER_Mobil17 = Beta('INTER_Mobil17', 0, None, None, 0)

B_Envir01_F1 = Beta('B_Envir01_F1', -1, None, None, 1)
B_Envir02_F1 = Beta('B_Envir02_F1', -1, None, None, 0)
B_Envir03_F1 = Beta('B_Envir03_F1', 1, None, None, 0)
B_Mobil11_F1 = Beta('B_Mobil11_F1', 1, None, None, 0)
B_Mobil14_F1 = Beta('B_Mobil14_F1', 1, None, None, 0)
B_Mobil16_F1 = Beta('B_Mobil16_F1', 1, None, None, 0)
B_Mobil17_F1 = Beta('B_Mobil17_F1', 1, None, None, 0)

MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

SIGMA_STAR_Envir01 = Beta('SIGMA_STAR_Envir01', 100, 1.0e-5, None, 0)
SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02', 0.01, 1.0e-5, None, 0)
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03', 100, 1.0e-5, None, 0)
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11', 100, 1.0e-5, None, 0)
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14', 100, 1.0e-5, None, 0)
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16', 100, 1.0e-5, None, 0)
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17', 100, 1.0e-5, None, 0)


delta_1 = Beta('delta_1', 0.1, 1.0e-5, None, 0)
delta_2 = Beta('delta_2', 0.2, 1.0e-5, None, 0)
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1
tau_3 = delta_1
tau_4 = delta_1 + delta_2

Envir01_tau_1 = (tau_1 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_2 = (tau_2 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_3 = (tau_3 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_4 = (tau_4 - MODEL_Envir01) / SIGMA_STAR_Envir01
IndEnvir01 = {
    1: bioNormalCdf(Envir01_tau_1),
    2: bioNormalCdf(Envir01_tau_2) - bioNormalCdf(Envir01_tau_1),
    3: bioNormalCdf(Envir01_tau_3) - bioNormalCdf(Envir01_tau_2),
    4: bioNormalCdf(Envir01_tau_4) - bioNormalCdf(Envir01_tau_3),
    5: 1 - bioNormalCdf(Envir01_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir01 = Elem(IndEnvir01, Envir01)

Envir02_tau_1 = (tau_1 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_2 = (tau_2 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_3 = (tau_3 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_4 = (tau_4 - MODEL_Envir02) / SIGMA_STAR_Envir02
IndEnvir02 = {
    1: bioNormalCdf(Envir02_tau_1),
    2: bioNormalCdf(Envir02_tau_2) - bioNormalCdf(Envir02_tau_1),
    3: bioNormalCdf(Envir02_tau_3) - bioNormalCdf(Envir02_tau_2),
    4: bioNormalCdf(Envir02_tau_4) - bioNormalCdf(Envir02_tau_3),
    5: 1 - bioNormalCdf(Envir02_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir02 = Elem(IndEnvir02, Envir02)

Envir03_tau_1 = (tau_1 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_2 = (tau_2 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_3 = (tau_3 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_4 = (tau_4 - MODEL_Envir03) / SIGMA_STAR_Envir03
IndEnvir03 = {
    1: bioNormalCdf(Envir03_tau_1),
    2: bioNormalCdf(Envir03_tau_2) - bioNormalCdf(Envir03_tau_1),
    3: bioNormalCdf(Envir03_tau_3) - bioNormalCdf(Envir03_tau_2),
    4: bioNormalCdf(Envir03_tau_4) - bioNormalCdf(Envir03_tau_3),
    5: 1 - bioNormalCdf(Envir03_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir03 = Elem(IndEnvir03, Envir03)

Mobil11_tau_1 = (tau_1 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_2 = (tau_2 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_3 = (tau_3 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_4 = (tau_4 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
IndMobil11 = {
    1: bioNormalCdf(Mobil11_tau_1),
    2: bioNormalCdf(Mobil11_tau_2) - bioNormalCdf(Mobil11_tau_1),
    3: bioNormalCdf(Mobil11_tau_3) - bioNormalCdf(Mobil11_tau_2),
    4: bioNormalCdf(Mobil11_tau_4) - bioNormalCdf(Mobil11_tau_3),
    5: 1 - bioNormalCdf(Mobil11_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil11 = Elem(IndMobil11, Mobil11)

Mobil14_tau_1 = (tau_1 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_2 = (tau_2 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_3 = (tau_3 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_4 = (tau_4 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
IndMobil14 = {
    1: bioNormalCdf(Mobil14_tau_1),
    2: bioNormalCdf(Mobil14_tau_2) - bioNormalCdf(Mobil14_tau_1),
    3: bioNormalCdf(Mobil14_tau_3) - bioNormalCdf(Mobil14_tau_2),
    4: bioNormalCdf(Mobil14_tau_4) - bioNormalCdf(Mobil14_tau_3),
    5: 1 - bioNormalCdf(Mobil14_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil14 = Elem(IndMobil14, Mobil14)

Mobil16_tau_1 = (tau_1 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_2 = (tau_2 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_3 = (tau_3 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_4 = (tau_4 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
IndMobil16 = {
    1: bioNormalCdf(Mobil16_tau_1),
    2: bioNormalCdf(Mobil16_tau_2) - bioNormalCdf(Mobil16_tau_1),
    3: bioNormalCdf(Mobil16_tau_3) - bioNormalCdf(Mobil16_tau_2),
    4: bioNormalCdf(Mobil16_tau_4) - bioNormalCdf(Mobil16_tau_3),
    5: 1 - bioNormalCdf(Mobil16_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil16 = Elem(IndMobil16, Mobil16)

Mobil17_tau_1 = (tau_1 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_2 = (tau_2 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_3 = (tau_3 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_4 = (tau_4 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
IndMobil17 = {
    1: bioNormalCdf(Mobil17_tau_1),
    2: bioNormalCdf(Mobil17_tau_2) - bioNormalCdf(Mobil17_tau_1),
    3: bioNormalCdf(Mobil17_tau_3) - bioNormalCdf(Mobil17_tau_2),
    4: bioNormalCdf(Mobil17_tau_4) - bioNormalCdf(Mobil17_tau_3),
    5: 1 - bioNormalCdf(Mobil17_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil17 = Elem(IndMobil17, Mobil17)

simulate = {
    'P_Envir01': P_Envir01,
    'P_Envir02': P_Envir02,
    'P_Envir03': P_Envir03,
    'P_Mobil11': P_Mobil11,
    'P_Mobil14': P_Mobil14,
    'P_Mobil16': P_Mobil16,
    'P_Mobil17': P_Mobil17,
}

beta_values = {}
for expr in simulate.values():
    beta_values = beta_values | expr.get_beta_values()

biosim = bio.BIOGEME(database, simulate)
biosim.modelName = '07problem_simul'

simulatedValues = biosim.simulate(theBetaValues=beta_values)
print(
    'We identify the entries for which the likelihood is zero, '
    'so that the log likelihood cannot be computed'
)
zero_values = simulatedValues.where(simulatedValues == 0, other='')
print(zero_values)
print('The problematic model is Envir02')



"""Sample alternatives

:author: Michel Bierlaire
:date: Sat Apr 15 16:01:41 2023

Generate a file with with choice data using:

    - a file containing the description of all alternatives,
          associated with an ID,
    - a file containing observations of individuals, associated with
          socio-economic characteristics, and an observed choice.

The choice must be a valid ID in the file of alternatives.  In the
choice data, a set of unchosen alternatives is sampled for each
individual observation using importance sampling. The correction term
for the model estimation is calculated and introduced in the file as
well.

In the generated choice data, alternative 0 is always the chosen one.
The name of the attributes of the alternatives are the names from the
original file containnig all alternatives, appended with the ID of the
alternative in the sample.

For instance, if there are 3 alternatives sampled per individual, and
the original fime contains the attributed "price", the choice data
will contain the columns "price_0", "price_1" and "price_2".

"""

import numpy as np
import pandas as pd
from biogeme.sampling import StratumTuple, sampling_of_alternatives

SAMPLE_SIZE = 10
FILE_NAME = 'sampled_alternatives.dat'

alternatives = pd.read_csv('restaurants.dat')
ID_COLUMN = 'ID'

observations = pd.read_csv('obs_choice.dat')
CHOICE_COLUMN = 'choice'

# Set of Asian restaurants
asian = set(alternatives[alternatives['Asian'] == 1][ID_COLUMN])

# Set of restaurants located in downtown
downtown = set(alternatives[alternatives['downtown'] == 1][ID_COLUMN])

# Set with all restaurants
all_alternatives = set(list(alternatives[ID_COLUMN]))

# Set of Asian restaurants in downtown
asian_and_downtown = asian & downtown

# Set of Asian restaurants, and of restaurants in downtown
asian_or_downtown = asian | downtown

# Set of Asian restaurants not in downtown
only_asian = asian - asian_and_downtown

# Set of non Asian restaurants in downtown
only_downtown = downtown - asian_and_downtown

# Set of restaurants that are neither Asian nor in downtown
others = all_alternatives - asian_or_downtown

half = int(np.floor(SAMPLE_SIZE / 2))
quarter = int(np.floor(SAMPLE_SIZE / 4))

# Importance sample.
# About 25% of the sample are Asian restaurants in downtown
# About 25% of the sample are Asian restaurants not in downtown
# About 25% of the sample are non Asian restaurants in downtown
# About 25% of the sample are meither Asian not in downtown
partition_asian_downtown = (
    StratumTuple(subset=asian_and_downtown, sample_size=quarter),
    StratumTuple(subset=only_asian, sample_size=quarter),
    StratumTuple(subset=only_downtown, sample_size=quarter),
    StratumTuple(subset=others, sample_size=SAMPLE_SIZE - 3 * quarter),
)

# If the script is executed, the alternatives are actually sampled. If
# it is imported into another script, nothing is actually done.

if __name__ == '__main__':
    sample = sampling_of_alternatives(
        partition=partition_asian_downtown,
        individuals=observations,
        choice_column=CHOICE_COLUMN,
        alternatives=alternatives,
        id_column=ID_COLUMN,
        always_include_chosen=True,
    )

    sample.to_csv(
        FILE_NAME,
        index=False,
    )

    print(f'File {FILE_NAME} has been generated')



"""Class that provides the function to the optimization algorithm

:author: Michel Bierlaire
:date: Wed Nov 30 10:17:26 2022

"""
from biogeme_optimization.function import FunctionToMinimize, FunctionData
import biogeme.exceptions as excep


class NegativeLikelihood(FunctionToMinimize):
    """Provides the value of the function to be minimized, as well as its
    derivatives. To be used by the opimization package.

    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, dimension, like, like_deriv, parameters=None):
        """Constructor"""

        tolerance = None
        steptol = None
        if parameters is not None:
            if 'tolerance' in parameters:
                tolerance = parameters['tolerance']
            if 'steptol' in parameters:
                steptol = parameters['steptol']

        super().__init__(epsilon=tolerance, steptol=steptol)

        self.the_dimension = dimension  #: number of parameters to estimate

        self.like = like  #: function calculating the log likelihood

        self.like_deriv = like_deriv
        """function calculating the log likelihood and its derivatives.
        """

    def dimension(self):
        """Provides the number of variables of the problem"""
        return self.the_dimension

    def _f(self):
        if self.x is None:
            raise excep.BiogemeError('The variables must be set first.')

        return -self.like(self.x, scaled=False, batch=None)

    def _f_g(self):
        if self.x is None:
            raise excep.BiogemeError('The variables must be set first.')

        f, g, *_ = self.like_deriv(
            self.x, scaled=False, hessian=False, bhhh=False, batch=None
        )

        return FunctionData(
            function=-f,
            gradient=-g,
            hessian=None,
        )

    def _f_g_h(self):
        if self.x is None:
            raise excep.BiogemeError('The variables must be set first.')

        f, g, h, _ = self.like_deriv(
            self.x, scaled=False, hessian=True, bhhh=False, batch=None
        )
        return FunctionData(
            function=-f,
            gradient=-g,
            hessian=-h,
        )



""" File: b07estimation_monte_carlo.py

 Author: Michel Bierlaire, EPFL
 Date: Thu Apr 13 21:04:47 2023

Estimation of a mixtures of logit models where the integral is
approximated using MonteCarlo integration.

"""

import biogeme.logging as blog
from biogeme.expressions import bioDraws
from b07estimation_specification import get_biogeme

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b07estimation_monte_carlo.py')

R = 2000
the_draws = bioDraws('B_TIME_RND', 'NORMAL')
the_biogeme = get_biogeme(the_draws)
the_biogeme.number_of_draws = R
the_biogeme.modelName = 'b07estimation_monte_carlo'
results = the_biogeme.estimate()

# Get the results in a pandas table
print(results.shortSummary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b06_point_elesticities.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:01:41 2023

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We calculate disaggregate and aggregate direct point elasticities.
"""

import sys
import biogeme.biogeme as bio
from biogeme import models
import biogeme.results as res
import biogeme.exceptions as excep
from biogeme.expressions import Derive
from optima_data import database, normalized_weight
from scenarios import (
    scenario,
    TimePT,
    TimeCar,
    MarginalCostPT,
    CostCarCHF,
    distance_km,
)


# Obtain the specification for the default scenario
V, nests, _, _ = scenario()

# Obtain the expression for the choice probability of each alternative
prob_PT = models.nested(V, None, nests, 0)
prob_CAR = models.nested(V, None, nests, 1)
prob_SM = models.nested(V, None, nests, 2)

# Calculation of the direct elasticities.
# We use the 'Derive' operator to calculate the derivatives.

direct_elas_pt_time = Derive(prob_PT, 'TimePT') * TimePT / prob_PT

direct_elas_pt_cost = Derive(prob_PT, 'MarginalCostPT') * MarginalCostPT / prob_PT

direct_elas_car_time = Derive(prob_CAR, 'TimeCar') * TimeCar / prob_CAR

direct_elas_car_cost = Derive(prob_CAR, 'CostCarCHF') * CostCarCHF / prob_CAR

direct_elas_sm_dist = Derive(prob_SM, 'distance_km') * distance_km / prob_SM

# Simulate the formulas
simulate = {
    'weight': normalized_weight,
    'Prob. car': prob_CAR,
    'Prob. public transportation': prob_PT,
    'Prob. slow modes': prob_SM,
    'direct_elas_pt_time': direct_elas_pt_time,
    'direct_elas_pt_cost': direct_elas_pt_cost,
    'direct_elas_car_time': direct_elas_car_time,
    'direct_elas_car_cost': direct_elas_car_cost,
    'direct_elas_sm_dist': direct_elas_sm_dist,
}

the_biogeme = bio.BIOGEME(database, simulate)

# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02estimation.py in order to generate '
        'the file b02estimation.pickle.'
    )

# simulatedValues is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulated_values = the_biogeme.simulate(results.getBetaValues())

# We calculate the aggregate elasticities

# First, the weighted probabilities
simulated_values['Weighted prob. car'] = (
    simulated_values['weight'] * simulated_values['Prob. car']
)
simulated_values['Weighted prob. PT'] = (
    simulated_values['weight'] * simulated_values['Prob. public transportation']
)
simulated_values['Weighted prob. SM'] = (
    simulated_values['weight'] * simulated_values['Prob. slow modes']
)

# Then the denominator of the aggregate elasticity expression.
denominator_car = simulated_values['Weighted prob. car'].sum()
denominator_pt = simulated_values['Weighted prob. PT'].sum()
denominator_sm = simulated_values['Weighted prob. SM'].sum()

# And finally the aggregate elasticities themselves.
direct_elas_term_car_time = (
    simulated_values['Weighted prob. car']
    * simulated_values['direct_elas_car_time']
    / denominator_car
).sum()
print(
    f'Aggregate direct point elasticity of car wrt time: '
    f'{direct_elas_term_car_time:.3g}'
)

direct_elas_term_car_cost = (
    simulated_values['Weighted prob. car']
    * simulated_values['direct_elas_car_cost']
    / denominator_car
).sum()
print(
    f'Aggregate direct point elasticity of car wrt cost: '
    f'{direct_elas_term_car_cost:.3g}'
)

direct_elas_term_pt_time = (
    simulated_values['Weighted prob. PT']
    * simulated_values['direct_elas_pt_time']
    / denominator_pt
).sum()
print(
    f'Aggregate direct point elasticity of PT wrt time: '
    f'{direct_elas_term_pt_time:.3g}'
)

direct_elas_term_pt_cost = (
    simulated_values['Weighted prob. PT']
    * simulated_values['direct_elas_pt_cost']
    / denominator_pt
).sum()
print(
    f'Aggregate direct point elasticity of PT wrt cost: '
    f'{direct_elas_term_pt_cost:.3g}'
)

direct_elas_term_sm_dist = (
    simulated_values['Weighted prob. SM']
    * simulated_values['direct_elas_sm_dist']
    / denominator_sm
).sum()
print(
    f'Aggregate direct point elasticity of SM wrt distance: '
    f'{direct_elas_term_sm_dist:.3g}'
)



"""File b13panel_simul.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:16:29 2023

 Example of a mixture of logit models, using Monte-Carlo integration.
 The datafile is organized as panel data.
 Three alternatives: Train, Car and Swissmetro

"""

import sys
import pickle
import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
import biogeme.exceptions as excep
import biogeme.results as res
from biogeme.expressions import (
    Beta,
    bioDraws,
    PanelLikelihoodTrajectory,
    MonteCarlo,
    log,
)
from swissmetro_panel import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

NUMBER_OF_DRAWS = 1000

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b13panel_simul.py')

# Parameters to be estimated
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME = Beta('B_TIME', 0, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL_ANTI')

# We do the same for the constants, to address serial correlation.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_CAR_S = Beta('ASC_CAR_S', 1, None, None, 0)
ASC_CAR_RND = ASC_CAR + ASC_CAR_S * bioDraws('ASC_CAR_RND', 'NORMAL_ANTI')

ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_TRAIN_S = Beta('ASC_TRAIN_S', 1, None, None, 0)
ASC_TRAIN_RND = ASC_TRAIN + ASC_TRAIN_S * bioDraws('ASC_TRAIN_RND', 'NORMAL_ANTI')

ASC_SM = Beta('ASC_SM', 0, None, None, 1)
ASC_SM_S = Beta('ASC_SM_S', 1, None, None, 0)
ASC_SM_RND = ASC_SM + ASC_SM_S * bioDraws('ASC_SM_RND', 'NORMAL_ANTI')

# Definition of the utility functions
V1 = ASC_TRAIN_RND + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM_RND + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR_RND + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to the random parameters, the likelihood of one observation is
# given by the logit model (called the kernel)
obsprob = models.logit(V, av, CHOICE)

# Conditional to the random parameters, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = PanelLikelihoodTrajectory(obsprob)

# We integrate over the random parameters using Monte-Carlo
logprob = log(MonteCarlo(condprobIndiv))

# Estimate the parameters.
try:
    results = res.bioResults(pickleFile='b12panel.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b12panel.py '
        'in order to generate the '
        'file b12panel.pickle.'
    )

# Simulate to recalculate the log likelihood directly from the
# formula, without the Biogeme object
simulated_loglike = logprob.getValue_c(
    database=database,
    betas=results.getBetaValues(),
    numberOfDraws=NUMBER_OF_DRAWS,
    aggregation=True,
    prepareIds=True,
)

print(f'Simulated log likelihood: {simulated_loglike}')

numerator = MonteCarlo(B_TIME_RND * condprobIndiv)
denominator = MonteCarlo(condprobIndiv)
simulate = {
    'Numerator': numerator,
    'Denominator': denominator,
}

biosim = bio.BIOGEME(database, simulate)
sim = biosim.simulate(results.getBetaValues())

sim['Individual-level parameters'] = sim['Numerator'] / sim['Denominator']
PICKLE_FILE = 'b13panel_individual_parameters.pickle'
with open(PICKLE_FILE, 'wb') as f:
    pickle.dump(sim, f)

HTML_FILE = 'b13panel_simul.html'
with open(HTML_FILE, 'w', encoding='utf-8') as h:
    print(sim.to_html(), file=h)
print(f'Simulated values available in {HTML_FILE}')



"""File b23binary_logit.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 17:58:18 2023

 Example of a binary logit model.
 Two alternatives: Train and Car
"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from swissmetro_binary import (
    database,
    CHOICE,
    TRAIN_AV_SP,
    CAR_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
B_TIME_CAR = Beta('B_TIME_CAR', 0, None, None, 0)
B_TIME_TRAIN = Beta('B_TIME_TRAIN', 0, None, None, 0)
B_COST_CAR = Beta('B_COST_CAR', 0, None, None, 0)
B_COST_TRAIN = Beta('B_COST_TRAIN', 0, None, None, 0)

# Definition of the utility functions
# We estimate a binary logit model. There are only two alternatives.
V1 = B_TIME_TRAIN * TRAIN_TT_SCALED + B_COST_TRAIN * TRAIN_COST_SCALED
V3 = ASC_CAR + B_TIME_CAR * CAR_TT_SCALED + B_COST_CAR * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b23logit'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b21multiple_models.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 16:58:49 2023

 Example of the estimation of several versions of the model using
 assisted specification algorithm Three alternatives: Train, Car and

"""
import biogeme.logging as blog
from biogeme.results import compile_estimation_results
from biogeme.multiobjectives import loglikelihood_dimension
from biogeme.assisted import AssistedSpecification
from b21multiple_models_spec import the_biogeme, PARETO_FILE_NAME

logger = blog.get_screen_logger(blog.INFO)
logger.info('Example b21multipleModels')

assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=loglikelihood_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

non_dominated_models = assisted_specification.run()

summary, description = compile_estimation_results(
    non_dominated_models, use_short_names=True
)
print(summary)
for k, v in description.items():
    if k != v:
        print(f'{k}: {v} AIC={summary.at["Akaike Information Criterion", k]}')



"""File everything_spec.py

:author: Michel Bierlaire, EPFL
:date: Sat Jul 15 15:40:33 2023

We investigate various specifications:
- 3 models
    - logit
    - nested logit with two nests: public and private transportation
    - nested logit with two nests existing and future modes
- 3 functional form for the travel time variables
    - linear specification,
    - Box-Cox transform,
    - power series,
- 2 specification for the cost coefficients:
    - generic
    - alternative specific
- 2 specification for the travel time coefficients:
    - generic
    - alternative specific
- 4 segmentations for the constants:
    - not segmented
    - segmented by GA (yearly subscription to public transport)
    - segmented by luggage
    - segmented both by GA and luggage
-  3 segmentations for the time coefficients:
    - not segmented
    - segmented with first class
    - segmented with trip purpose

This leads to a total of 432 specifications.
"""
import numpy as np
from biogeme import models
from biogeme.expressions import Beta
from biogeme.catalog import (
    Catalog,
    segmentation_catalogs,
    generic_alt_specific_catalogs,
)

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

segmentation_ga = database.generate_segmentation(
    variable='GA', mapping={0: 'noGA', 1: 'GA'}
)

segmentation_luggage = database.generate_segmentation(
    variable='LUGGAGE', mapping={0: 'no_lugg', 1: 'one_lugg', 3: 'several_lugg'}
)

segmentation_first = database.generate_segmentation(
    variable='FIRST', mapping={0: '2nd_class', 1: '1st_class'}
)

# We consider two trip purposes: 'commuters' and anything else. We
# need to define a binary variable first

database.data['COMMUTERS'] = np.where(database.data['PURPOSE'] == 1, 1, 0)

segmentation_purpose = database.generate_segmentation(
    variable='COMMUTERS', mapping={0: 'non_commuters', 1: 'commuters'}
)


# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Non linear specifications for the travel time

# Parameter of the Box-Cox transform
ell_travel_time = Beta('lambda_travel_time', 1, None, 10, 0)

# Coefficients of the power series
square_tt_coef = Beta('square_tt_coef', 0, None, None, 0)
cube_tt_coef = Beta('cube_tt_coef', 0, None, None, 0)


def power_series(the_variable):
    """Generate the expression of a polynomial of degree 3

    :param the_variable: variable of the polynomial
    :type the_variable: biogeme.expressions.Expression
    """
    return (
        the_variable
        + square_tt_coef * the_variable**2
        + cube_tt_coef * the_variable * the_variable**3
    )


linear_train_tt = TRAIN_TT_SCALED
boxcox_train_tt = models.boxcox(TRAIN_TT_SCALED, ell_travel_time)
power_train_tt = power_series(TRAIN_TT_SCALED)
train_tt_catalog = Catalog.from_dict(
    catalog_name='train_tt_catalog',
    dict_of_expressions={
        'linear': linear_train_tt,
        'boxcox': boxcox_train_tt,
        'power': power_train_tt,
    },
)

linear_sm_tt = SM_TT_SCALED
boxcox_sm_tt = models.boxcox(SM_TT_SCALED, ell_travel_time)
power_sm_tt = power_series(SM_TT_SCALED)
sm_tt_catalog = Catalog.from_dict(
    catalog_name='sm_tt_catalog',
    dict_of_expressions={
        'linear': linear_sm_tt,
        'boxcox': boxcox_sm_tt,
        'power': power_sm_tt,
    },
    controlled_by=train_tt_catalog.controlled_by,
)

linear_car_tt = CAR_TT_SCALED
boxcox_car_tt = models.boxcox(CAR_TT_SCALED, ell_travel_time)
power_car_tt = power_series(CAR_TT_SCALED)

car_tt_catalog = Catalog.from_dict(
    catalog_name='car_tt_catalog',
    dict_of_expressions={
        'linear': linear_car_tt,
        'boxcox': boxcox_car_tt,
        'power': power_car_tt,
    },
    controlled_by=train_tt_catalog.controlled_by,
)


ASC_TRAIN_catalog, ASC_CAR_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_TRAIN, ASC_CAR],
    potential_segmentations=(
        segmentation_ga,
        segmentation_luggage,
    ),
    maximum_number=2,
)


(B_TIME_catalog,) = generic_alt_specific_catalogs(
    generic_name='B_TIME',
    beta_parameters=[B_TIME],
    alternatives=['TRAIN', 'SM', 'CAR'],
    potential_segmentations=(
        segmentation_first,
        segmentation_purpose,
    ),
    maximum_number=1,
)

(B_COST_catalog,) = generic_alt_specific_catalogs(
    generic_name='B_COST', beta_parameters=[B_COST], alternatives=['TRAIN', 'SM', 'CAR']
)

# Definition of the utility functions
V1 = (
    ASC_TRAIN_catalog
    + B_TIME_catalog['TRAIN'] * train_tt_catalog
    + B_COST_catalog['TRAIN'] * TRAIN_COST_SCALED
)
V2 = B_TIME_catalog['SM'] * sm_tt_catalog + B_COST_catalog['SM'] * SM_COST_SCALED
V3 = (
    ASC_CAR_catalog
    + B_TIME_catalog['CAR'] * car_tt_catalog
    + B_COST_catalog['CAR'] * CAR_CO_SCALED
)

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob_logit = models.loglogit(V, av, CHOICE)

MU_existing = Beta('MU_existing', 1, 1, 10, 0)
existing = MU_existing, [1, 3]
future = 1.0, [2]
nests_existing = existing, future
logprob_nested_existing = models.lognested(V, av, nests_existing, CHOICE)

MU_public = Beta('MU_public', 1, 1, 10, 0)
public = MU_public, [1, 2]
private = 1.0, [3]
nests_public = public, private
logprob_nested_public = models.lognested(V, av, nests_public, CHOICE)

model_catalog = Catalog.from_dict(
    catalog_name='model_catalog',
    dict_of_expressions={
        'logit': logprob_logit,
        'nested existing': logprob_nested_existing,
        'nested public': logprob_nested_public,
    },
)



"""
Test the biogeme module

:author: Michel Bierlaire
:date: Wed Apr 29 18:32:42 2020

"""
# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring

import os
import unittest
import random as rnd
import numpy as np
import biogeme.biogeme as bio
from biogeme.catalog import Catalog
import biogeme.exceptions as excep
from biogeme.expressions import (
    Variable,
    Beta,
    bioDraws,
    MonteCarlo,
    RandomVariable,
    PanelLikelihoodTrajectory,
    Numeric,
)
from test_data import getData, getPanelData


class TestBiogeme(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(90267)
        rnd.seed(90267)

    @classmethod
    def tearDownClass(cls):
        for file_name in os.listdir('.'):
            if (
                file_name.endswith('.pickle')
                or file_name.endswith('.html')
                or file_name.endswith('.iter')
                or file_name.endswith('.log')
            ):
                os.remove(file_name)

    def get_dict_of_expressions(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, -3, 3, 0)
        beta2 = Beta('beta2', 2.0, -3, 10, 0)
        likelihood = -((beta1 * Variable1) ** 2) - (beta2 * Variable2) ** 2
        simul = (beta1 + 2 * beta2) / Variable1 + (beta2 + 2 * beta1) / Variable2
        dict_of_expressions = {
            'loglike': likelihood,
            'weight': Numeric(1),
            'beta1': beta1,
            'simul': simul,
        }
        return dict_of_expressions

    def get_biogeme_instance(self):
        data = getData(1)
        return bio.BIOGEME(data, self.get_dict_of_expressions())

    def get_biogeme_instance_without_bounds(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, None, None, 0)
        beta2 = Beta('beta2', 2.0, None, None, 0)
        likelihood = -((beta1 * Variable1) ** 2) - (beta2 * Variable2) ** 2
        simul = (beta1 + 2 * beta2) / Variable1 + (beta2 + 2 * beta1) / Variable2
        dict_of_expressions = {
            'loglike': likelihood,
            'weight': Numeric(1),
            'beta1': beta1,
            'simul': simul,
        }
        data = getData(1)
        return bio.BIOGEME(data, dict_of_expressions)

    def get_panel_instance(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, -3, 3, 0)
        beta2 = Beta('beta2', 2.0, -3, 10, 0)
        likelihood = -((beta1 * Variable1) ** 2) - (beta2 * Variable2) ** 2
        simul = (beta1 + 2 * beta2) / Variable1 + (beta2 + 2 * beta1) / Variable2
        dict_of_expressions = {
            'loglike': likelihood,
            'weight': Numeric(1),
            'beta1': beta1,
            'simul': simul,
        }
        data = getData(1)
        data.panel('Person')
        return bio.BIOGEME(data, dict_of_expressions)

    def test_ctor(self):
        # Test obsolete parameters
        aBiogeme = bio.BIOGEME(
            getData(1),
            self.get_dict_of_expressions(),
        )
        wrong_data = getData(1)
        wrong_data.data.loc['Person', 0] = np.nan
        with self.assertRaises(excep.BiogemeError):
            bBiogeme = bio.BIOGEME(
                wrong_data,
                self.get_dict_of_expressions(),
            )

        with self.assertRaises(excep.BiogemeError):
            bBiogeme = bio.BIOGEME(
                getData(1),
                'wrong_object',
            )

        with self.assertRaises(excep.BiogemeError):
            bBiogeme = bio.BIOGEME(
                getData(1),
                {'loglike': 'wrong_object'},
            )

        wrong_expression = Variable('Variable1') * PanelLikelihoodTrajectory(
            Beta('beta1', -1.0, -3, 3, 0)
        )
        with self.assertRaises(excep.BiogemeError):
            bBiogeme = bio.BIOGEME(
                getPanelData(1),
                wrong_expression,
            )

        cBiogeme = bio.BIOGEME(
            getPanelData(1),
            self.get_dict_of_expressions(),
        )

    def test_old_parameters(self):
        b1 = bio.BIOGEME(getData(1), self.get_dict_of_expressions(), numberOfDraws=12)
        self.assertEqual(b1.number_of_draws, 12)

        b2 = bio.BIOGEME(getData(1), self.get_dict_of_expressions(), numberOfThreads=12)
        self.assertEqual(b2.number_of_threads, 12)

        b3 = bio.BIOGEME(getData(1), self.get_dict_of_expressions(), seed=12)
        self.assertEqual(b3.seed_param, 12)

        b4 = bio.BIOGEME(getData(1), self.get_dict_of_expressions(), missingData=12)
        self.assertEqual(b4.missing_data, 12)

        b5 = bio.BIOGEME(getData(1), self.get_dict_of_expressions(), suggestScales=True)

    def test_saved_iterations(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme._loadSavedIteration()
        # Remove the file and try to loaed it again
        my_biogeme._loadSavedIteration()

    def test_random_init_values(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.setRandomInitValues(defaultBound=10)
        for v in my_biogeme.id_manager.free_betas_values:
            self.assertLessEqual(v, 10)
            self.assertGreaterEqual(v, -10)

    def test_parameters(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.algorithm_name = 'scipy'
        self.assertEqual(my_biogeme.algorithm_name, 'scipy')

        my_biogeme.identification_threshold = 1.0e-5
        self.assertEqual(my_biogeme.identification_threshold, 1.0e-5)

        my_biogeme.seed_param = 123
        self.assertEqual(my_biogeme.seed_param, 123)

        my_biogeme.save_iterations = False
        self.assertEqual(my_biogeme.save_iterations, False)

        my_biogeme.saveIterations = False
        self.assertEqual(my_biogeme.saveIterations, False)

        my_biogeme.missing_data = 921967
        self.assertEqual(my_biogeme.missing_data, 921967)

        my_biogeme.missingData = 921967
        self.assertEqual(my_biogeme.missingData, 921967)

        my_biogeme.number_of_threads = 921967
        self.assertEqual(my_biogeme.number_of_threads, 921967)

        my_biogeme.number_of_draws = 921967
        self.assertEqual(my_biogeme.number_of_draws, 921967)

        my_biogeme.numberOfDraws = 921967
        self.assertEqual(my_biogeme.numberOfDraws, 921967)

        my_biogeme.only_robust_stats = True
        self.assertEqual(my_biogeme.only_robust_stats, True)

        my_biogeme.generate_html = False
        self.assertEqual(my_biogeme.generate_html, False)

        my_biogeme.generate_html = False
        self.assertEqual(my_biogeme.generate_html, False)

        my_biogeme.generate_pickle = False
        self.assertEqual(my_biogeme.generate_pickle, False)

        my_biogeme.generate_pickle = False
        self.assertEqual(my_biogeme.generate_pickle, False)

        my_biogeme.tolerance = 1.0e-5
        self.assertEqual(my_biogeme.tolerance, 1.0e-5)

        my_biogeme.second_derivatives = 0.3
        self.assertEqual(my_biogeme.second_derivatives, 0.3)

        my_biogeme.infeasible_cg = True
        self.assertEqual(my_biogeme.infeasible_cg, True)

        my_biogeme.initial_radius = 3.14
        self.assertEqual(my_biogeme.initial_radius, 3.14)

        my_biogeme.steptol = 3.14
        self.assertEqual(my_biogeme.steptol, 3.14)

        my_biogeme.enlarging_factor = 3.14
        self.assertEqual(my_biogeme.enlarging_factor, 3.14)

        my_biogeme.maxiter = 314
        self.assertEqual(my_biogeme.maxiter, 314)

        my_biogeme.dogleg = True
        self.assertEqual(my_biogeme.dogleg, True)

    def test_free_beta_names(self):
        my_biogeme = self.get_biogeme_instance()
        result = my_biogeme.free_beta_names()
        expected_result = ['beta1', 'beta2']
        self.assertListEqual(result, expected_result)

    def test_bounds_on_beta(self):
        my_biogeme = self.get_biogeme_instance()
        result = my_biogeme.getBoundsOnBeta('beta2')
        expected_result = -3, 10
        self.assertTupleEqual(result, expected_result)
        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.getBoundsOnBeta('wrong_name')

    def test_saveIterationsFileName(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.generate_html = False
        my_biogeme.generate_pickle = False
        my_biogeme.saveIterations = False
        my_biogeme.modelName = 'simpleExample'
        f = my_biogeme._saveIterationsFileName()
        self.assertEqual(f, '__simpleExample.iter')

    def test_generateDraws(self):
        the_data = getData(1)
        self.assertIsNone(the_data.theDraws)
        ell = bioDraws('test', 'NORMAL')
        with self.assertRaises(excep.BiogemeError):
            b = bio.BIOGEME(the_data, ell)
        b = bio.BIOGEME(the_data, ell, skip_audit=True)
        b._generateDraws(10)
        self.assertTupleEqual(the_data.theDraws.shape, (5, 10, 1))
        ell2 = bioDraws('test', 'NORMAL') + bioDraws('test2', 'UNIFORM')
        b2 = bio.BIOGEME(the_data, ell2, skip_audit=True)
        b2._generateDraws(20)
        self.assertTupleEqual(the_data.theDraws.shape, (5, 20, 2))

    def test_random_variable(self):
        rv = RandomVariable('omega')
        with self.assertRaises(excep.BiogemeError):
            b = bio.BIOGEME(getData(1), rv)

    def test_getBoundsOnBeta(self):
        my_biogeme = self.get_biogeme_instance()
        b = my_biogeme.getBoundsOnBeta('beta1')
        self.assertTupleEqual(b, (-3, 3))
        b = my_biogeme.getBoundsOnBeta('beta2')
        self.assertTupleEqual(b, (-3, 10))

    def test_calculateNullLoglikelihood(self):
        my_biogeme = self.get_biogeme_instance()
        null_ell = my_biogeme.calculateNullLoglikelihood({1: 1, 2: 1})
        self.assertAlmostEqual(null_ell, -3.4657359027997265, 2)
        null_ell_2 = my_biogeme.calculateNullLoglikelihood({1: 1, 2: 1, 3: 1})
        self.assertAlmostEqual(null_ell_2, -5.493061443340549, 2)

    def test_calculateInitLikelihood(self):
        my_biogeme = self.get_biogeme_instance()
        res = my_biogeme.calculateInitLikelihood()
        self.assertAlmostEqual(res, -22055, 5)

    def test_calculateLikelihood(self):
        my_biogeme = self.get_biogeme_instance()
        x = my_biogeme.id_manager.free_betas_values
        xplus = [v + 1 for v in x]
        res = my_biogeme.calculateLikelihood(xplus, scaled=False)
        self.assertEqual(res, -49500)
        with self.assertRaises(ValueError):
            _ = my_biogeme.calculateLikelihood([1], scaled=False)

    def test_calculateLikelihoodAndDerivatives(self):
        my_biogeme = self.get_biogeme_instance()
        x = my_biogeme.id_manager.free_betas_values
        xplus = [v + 1 for v in x]
        my_biogeme.save_iterations = True
        f, g, h, bhhh = my_biogeme.calculateLikelihoodAndDerivatives(
            xplus, scaled=False, hessian=True, bhhh=True
        )
        f_true = -49500
        g_true = [0.0, -33000.0]
        h_true = [[0.0, 0.0], [0.0, -11000.0]]
        bhhh_true = [[0.0, 0.0], [0.0, 352440000.0]]
        self.assertEqual(f_true, f)
        self.assertListEqual(g_true, g.tolist())
        self.assertListEqual(h_true, h.tolist())
        self.assertListEqual(bhhh_true, bhhh.tolist())
        with self.assertRaises(ValueError):
            _ = my_biogeme.calculateLikelihoodAndDerivatives([1], scaled=False)
        my_biogeme.database.data = my_biogeme.database.data[0:0]
        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.calculateLikelihoodAndDerivatives(x, scaled=True)

    def test_likelihoodFiniteDifferenceHessian(self):
        my_biogeme = self.get_biogeme_instance()
        x = my_biogeme.id_manager.free_betas_values
        xplus = [v + 1 for v in x]
        h = my_biogeme.likelihoodFiniteDifferenceHessian(xplus)
        h_true = [[-110, 0.0], [0.0, -11000]]
        for row, row_true in zip(h, h_true):
            for col, col_true in zip(row, row_true):
                self.assertAlmostEqual(col, col_true, 3)

    def test_checkDerivatives(self):
        my_biogeme = self.get_biogeme_instance()
        _, _, _, gdiff, hdiff = my_biogeme.checkDerivatives([1, 10])
        gdiff_true = [-0.00032656, 0.00545755]
        hdiff_true = [[6.49379217e-09, 0.00000000e00], [0.00000000e00, -1.39698386e-06]]
        for col, col_true in zip(gdiff, gdiff_true):
            self.assertAlmostEqual(col, col_true, 3)
        for row, row_true in zip(hdiff, hdiff_true):
            for col, col_true in zip(row, row_true):
                self.assertAlmostEqual(col, col_true, 3)

    def test_estimate(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.generate_html = False
        my_biogeme.generate_pickle = False
        my_biogeme.saveIterations = False
        my_biogeme.modelName = 'simpleExample'
        my_biogeme.numberOfThreads = 1
        my_biogeme.algorithm_name = 'simple_bounds'
        my_biogeme.bootstrap_samples = 10
        results = my_biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        panelBiogeme = bio.BIOGEME(getPanelData(1), self.get_dict_of_expressions())
        panelBiogeme.bootstrap_samples = 10
        results = panelBiogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.data.logLike, 0, 5)

        my_biogeme.algorithm_name = 'scipy'
        my_biogeme.saveIterations = True
        results = my_biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        # We try to recycle, while there is no pickle file yet.
        results = my_biogeme.estimate(recycle=True)
        # We estimate the model twice to generate two pickle files.
        my_biogeme.generate_pickle = True
        results = my_biogeme.estimate()
        results = my_biogeme.estimate()
        results = my_biogeme.estimate(recycle=True)
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme.generate_pickle = False

        my_biogeme_without_bounds = self.get_biogeme_instance_without_bounds()
        my_biogeme_without_bounds.generate_html = False
        my_biogeme_without_bounds.generate_pickle = False
        my_biogeme_without_bounds.saveIterations = False

        my_biogeme_without_bounds.algorithm_name = 'TR-newton'
        results = my_biogeme_without_bounds.estimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme_without_bounds.algorithm_name = 'TR-BFGS'
        results = my_biogeme_without_bounds.estimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme_without_bounds.algorithm_name = 'LS-newton'
        results = my_biogeme_without_bounds.estimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme_without_bounds.algorithm_name = 'LS-BFGS'
        results = my_biogeme_without_bounds.estimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)

        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.estimate(algorithm='any_algo')
        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.estimate(algoParameters='any_param')

        aBiogeme = bio.BIOGEME(getData(1), {'loglike': Numeric(0)})
        with self.assertRaises(excep.BiogemeError):
            _ = aBiogeme.estimate()

        aBiogeme.loglike = None
        with self.assertRaises(excep.BiogemeError):
            _ = aBiogeme.estimate()

    def test_quickEstimate(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.generate_html = False
        my_biogeme.generate_pickle = False
        my_biogeme.saveIterations = False
        my_biogeme.modelName = 'simpleExample'
        my_biogeme.numberOfThreads = 1
        my_biogeme.algorithm_name = 'simple_bounds'
        my_biogeme.bootstrap_samples = 10
        results = my_biogeme.quickEstimate(run_bootstrap=True)
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        panelBiogeme = bio.BIOGEME(getPanelData(1), self.get_dict_of_expressions())
        panelBiogeme.bootstrap_samples = 10
        results = panelBiogeme.quickEstimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)

        my_biogeme.algorithm_name = 'scipy'
        my_biogeme.saveIterations = True
        results = my_biogeme.quickEstimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        # We try to recycle, while there is no pickle file yet.
        results = my_biogeme.quickEstimate(recycle=True)
        # We quickEstimate the model twice to generate two pickle files.
        my_biogeme.generate_pickle = True
        results = my_biogeme.quickEstimate()
        results = my_biogeme.quickEstimate()
        results = my_biogeme.quickEstimate(recycle=True)
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme.generate_pickle = False

        my_biogeme_without_bounds = self.get_biogeme_instance_without_bounds()
        my_biogeme_without_bounds.generate_html = False
        my_biogeme_without_bounds.generate_pickle = False
        my_biogeme_without_bounds.saveIterations = False

        my_biogeme_without_bounds.algorithm_name = 'TR-newton'
        results = my_biogeme_without_bounds.quickEstimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme_without_bounds.algorithm_name = 'TR-BFGS'
        results = my_biogeme_without_bounds.quickEstimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme_without_bounds.algorithm_name = 'LS-newton'
        results = my_biogeme_without_bounds.quickEstimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)
        my_biogeme_without_bounds.algorithm_name = 'LS-BFGS'
        results = my_biogeme_without_bounds.quickEstimate()
        self.assertAlmostEqual(results.data.logLike, 0, 5)

        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.quickEstimate(algorithm='any_algo')
        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.quickEstimate(algoParameters='any_param')

        aBiogeme = bio.BIOGEME(getData(1), {'loglike': Numeric(0)})
        with self.assertRaises(excep.BiogemeError):
            _ = aBiogeme.quickEstimate()

        aBiogeme.loglike = None
        with self.assertRaises(excep.BiogemeError):
            _ = aBiogeme.quickEstimate()

    def test_simulate(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.algorithm_name = 'simple_bounds'
        results = my_biogeme.estimate()
        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.simulate(theBetaValues=None)

        s = my_biogeme.simulate(results.getBetaValues())
        self.assertAlmostEqual(s.loc[0, 'loglike'], 0, 3)

        the_betas = results.getBetaValues()
        the_betas['any_beta'] = 0.1
        s = my_biogeme.simulate(the_betas)
        self.assertAlmostEqual(s.loc[0, 'loglike'], 0, 3)

        with self.assertRaises(excep.BiogemeError):
            _ = my_biogeme.simulate('wrong_object')

        myPanelBiogeme = bio.BIOGEME(getPanelData(1), self.get_dict_of_expressions())
        results = myPanelBiogeme.estimate()
        with self.assertRaises(excep.BiogemeError):
            s = myPanelBiogeme.simulate(results.getBetaValues())

        myPanelBiogeme = bio.BIOGEME(
            getPanelData(1),
            {
                'Simul': MonteCarlo(
                    PanelLikelihoodTrajectory(bioDraws('test', 'NORMAL'))
                )
            },
        )
        s = myPanelBiogeme.simulate(results.getBetaValues())

    def test_changeInitValues(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.change_init_values({'beta2': -100, 'beta1': 3.14156})
        self.assertListEqual(my_biogeme.id_manager.free_betas_values, [3.14156, -100])

    def test_confidenceIntervals(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.bootstrap_samples = 100
        results = my_biogeme.estimate(run_bootstrap=True)
        drawsFromBetas = results.getBetasForSensitivityAnalysis(
            my_biogeme.id_manager.free_betas.names
        )
        s = my_biogeme.simulate(results.getBetaValues())
        left, right = my_biogeme.confidenceIntervals(drawsFromBetas)
        self.assertLessEqual(left.loc[0, 'loglike'], s.loc[0, 'loglike'])
        self.assertGreaterEqual(right.loc[0, 'loglike'], s.loc[0, 'loglike'])

    def test_validate(self):
        my_data = getData(1)
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.algorithm_name = 'simple_bounds'
        results = my_biogeme.estimate()
        validation_data = my_data.split(slices=2)
        validation_results = my_biogeme.validate(results, validation_data)
        self.assertAlmostEqual(validation_results[0]['Loglikelihood'].sum(), 0, 3)
        self.assertAlmostEqual(validation_results[1]['Loglikelihood'].sum(), 0, 3)

        b = self.get_panel_instance()
        results = b.estimate()
        panel_data = getPanelData(1)
        validation_data = panel_data.split(slices=2)
        with self.assertRaises(excep.BiogemeError):
            validation_results = b.validate(results, validation_data)

    def test_optimize(self):
        # Here, we test only the special cases, as it has been called
        # several times by estimate
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.algorithm_name = 'simple_bounds'
        my_biogeme.optimize()
        my_biogeme._algorithm = None
        with self.assertRaises(excep.BiogemeError):
            my_biogeme.optimize()

    def test_print(self):
        my_biogeme = self.get_biogeme_instance()
        result = str(my_biogeme)[0:20]
        expected_result = 'biogemeModelDefaultN'
        self.assertEqual(result, expected_result)

    def test_files(self):
        my_biogeme = self.get_biogeme_instance()
        my_biogeme.algorithm_name = 'simple_bounds'
        my_biogeme.modelName = 'name_for_file'
        my_biogeme.generate_html = True
        my_biogeme.estimate()
        result = my_biogeme.files_of_type('html', all_files=False)
        expected_result = ['name_for_file.html']
        self.assertListEqual(result, expected_result)
        result = my_biogeme.files_of_type('html', all_files=True)
        self.assertGreaterEqual(len(result), 1)

    def test_from_configuration(self):
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        catalog = Catalog.from_dict(
            catalog_name='the_catalog',
            dict_of_expressions={
                'beta_1': beta_1,
                'beta_2': beta_2,
            },
        )

        # Valid constructor
        config_id = 'the_catalog:beta_2'
        the_biogeme = bio.BIOGEME.from_configuration(
            config_id=config_id,
            expression=catalog,
            database=getData(1),
        )
        self.assertIs(the_biogeme.loglike, catalog)

        # Invalid constructor: typo
        config_id = 'the_catalog:bta_2'
        with self.assertRaises(excep.BiogemeError):
            _ = bio.BIOGEME.from_configuration(
                config_id=config_id,
                expression=catalog,
                database=getData(1),
            )

        # Invalid constructor: wrong id
        config_id = 'wrong_id'
        with self.assertRaises(excep.BiogemeError):
            _ = bio.BIOGEME.from_configuration(
                config_id=config_id,
                expression=catalog,
                database=getData(1),
            )


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models

from spec_optima import V, av, Choice

myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/optima.dat', sep='\t')
df.loc[df['OccupStat'] > 2, 'OccupStat'] = 3
df.loc[df['OccupStat'] == -1, 'OccupStat'] = 3
df.loc[df['Education'] <= 3, 'Education'] = 3
df.loc[df['Education'] <= 3, 'Education'] = 3
df.loc[df['Education'] == 5, 'Education'] = 4
df.loc[df['Education'] == 8, 'Education'] = 7
df.loc[df['TripPurpose'] != 1, 'TripPurpose'] = 2
df.loc[df['CarAvail'] != 3, 'CarAvail'] = 1

exclude = (
    (df.Choice == -1)
    | (df.CostCarCHF < 0)
    | (df.CarAvail == 3) & (df.Choice == 1)
)
df.drop(df[exclude].index, inplace=True)
df = df.reset_index()

database = db.Database('optima', df)


class test_01(unittest.TestCase):
    def testQuickEstimate(self):
        logprob = models.loglogit(V, av, Choice)
        biogeme = bio.BIOGEME(database, logprob)
        biogeme.modelName = 'test_01'
        biogeme.generateHtml = False
        biogeme.generatePickle = False
        biogeme.saveIterations = False
        results = biogeme.quickEstimate()
        self.assertAlmostEqual(results.data.logLike,  -1068.78, 1)


if __name__ == '__main__':
    unittest.main()



"""
Test the specification module

:author: Michel Bierlaire
:date: Tue Jun 20 08:56:40 2023
"""

import unittest
from biogeme.specification import Specification
from spec_swissmetro import logprob
from swissmetro_data import database


class TestSpecification(unittest.TestCase):
    def setUp(self):
        self.expression = logprob
        Specification.expression = logprob
        Specification.database = database
        self.config_id_1 = 'ASC:no_seg'
        self.config_id_2 = 'ASC:MALE'

    def test_constructor(self):
        config = self.expression.current_configuration()
        the_specification = Specification(config)
        the_results = the_specification.get_results()
        expected_loglike = -5331.252006916163
        loglike = the_results.data.logLike
        self.assertAlmostEqual(loglike, expected_loglike, 2)

    def test_from_string_id(self):
        the_specification = Specification.from_string_id(self.config_id_2)
        the_results = the_specification.get_results()
        expected_loglike = -5187.983411661233
        loglike = the_results.data.logLike
        self.assertAlmostEqual(loglike, expected_loglike, 2)

    def test_default_specification(self):
        the_specification = Specification.default_specification()
        the_results = the_specification.get_results()
        expected_loglike = -5331.252006916163
        loglike = the_results.data.logLike
        self.assertAlmostEqual(loglike, expected_loglike, 2)

    def test_describe(self):
        the_specification = Specification.from_string_id(self.config_id_2)
        expected_result = 'Final log likelihood:		-5187.983'
        description = the_specification.describe()[102:134]
        self.assertEqual(description, expected_result)


if __name__ == '__main__':
    unittest.main()



"""
Test the sampling module

:author: Michel Bierlaire
:date: Sun Jan  8 18:03:42 2023
"""
# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring

import unittest
import numpy as np
import pandas as pd
import biogeme.sampling as sam
import biogeme.exceptions as excep


class TestSampling(unittest.TestCase):
    def setUp(self):
        self.nbr_of_alt = 20
        self.nbr_of_obs = 10
        self.alternatives = pd.DataFrame(
            {
                'ID': list(range(self.nbr_of_alt)),
                'Attr1': np.random.rand(self.nbr_of_alt),
            }
        )
        self.individuals = pd.DataFrame(
            {
                'Char1': np.random.randint(2, size=self.nbr_of_obs),
                'choice': np.random.randint(self.nbr_of_alt, size=self.nbr_of_obs),
            }
        )

    def test_sampling(self):
        set1 = set(range(10))
        set2 = set(range(10, 20))
        partition = (
            sam.StratumTuple(subset=set1, sample_size=3),
            sam.StratumTuple(subset=set2, sample_size=3),
        )
        the_sample = sam.sampling_of_alternatives(
            partition=partition,
            individuals=self.individuals,
            choice_column='choice',
            alternatives=self.alternatives,
            id_column='ID',
            always_include_chosen=True,
        )
        the_sample = the_sample.astype({f'ID_{i}': int for i in range(6)})
        pd.testing.assert_series_equal(
            the_sample['choice'], the_sample['ID_0'], check_names=False
        )

    def test_partition_1(self):
        set1 = set([1, 2, 3])
        set2 = set([3, 4, 5])
        partition = (
            sam.StratumTuple(subset=set1, sample_size=3),
            sam.StratumTuple(subset=set2, sample_size=3),
        )
        with self.assertRaises(excep.BiogemeError):
            the_sample = sam.sampling_of_alternatives(
                partition=partition,
                individuals=self.individuals,
                choice_column='choice',
                alternatives=self.alternatives,
                id_column='ID',
                always_include_chosen=True,
            )

    def test_partition_2(self):
        set1 = set([1, 2])
        set2 = set([4, 5])
        partition = (
            sam.StratumTuple(subset=set1, sample_size=3),
            sam.StratumTuple(subset=set2, sample_size=3),
        )
        with self.assertRaises(excep.BiogemeError):
            the_sample = sam.sampling_of_alternatives(
                partition=partition,
                individuals=self.individuals,
                choice_column='choice',
                alternatives=self.alternatives,
                id_column='ID',
                always_include_chosen=True,
            )

    def test_partition_3(self):
        set1 = set([1, 2, 3])
        set2 = set([4, 5, 6])
        partition = (
            sam.StratumTuple(subset=set1, sample_size=3),
            sam.StratumTuple(subset=set2, sample_size=3),
        )
        with self.assertRaises(excep.BiogemeError):
            the_sample = sam.sampling_of_alternatives(
                partition=partition,
                individuals=self.individuals,
                choice_column='choice',
                alternatives=self.alternatives,
                id_column='ID',
                always_include_chosen=True,
            )


if __name__ == '__main__':
    unittest.main()
